import numpy as np 
import tesuract
from collections import defaultdict
import time as T
import matplotlib.pyplot as mpl
import warnings, pdb
 
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.estimator_checks import check_estimator 
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import GridSearchCV, ParameterGrid, KFold, cross_val_score, cross_validate
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import FunctionTransformer
from sklearn.metrics import make_scorer
from sklearn.decomposition import PCA, FastICA
from sklearn.pipeline import Pipeline
from sklearn.neural_network import MLPRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR

from alive_progress import alive_bar
		
class RegressionWrapperCV(BaseEstimator):
	def __init__(self,regressor='pce',reg_params={},scorer='neg_root_mean_squared_error',n_jobs=1,verbose=1,cv=None):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.verbose = verbose
		self.cv = cv
		# self.__dict__.update(reg_params)
	def _setup_cv(self):
		if self.cv == None:
			self.cv = KFold(n_splits=5)
	def _reformat_grid(self,params):
		try: 
			# if parameter grid is in the form of grid search cv
			ParameterGrid(params)
			params_cv = params
		except:
			# convert reg_params to form that can be used by grid search
			# for a single grid point
			params_cv = defaultdict(list)
			for key,item in params.items():
				params_cv[key].append(item)
		return params_cv
	def _model_factory(self,regressor):
		if regressor == 'pce':
			return tesuract.PCEReg()
		if regressor == 'rf':
			return RandomForestRegressor()
		if regressor == 'mlp':
			return MLPRegressor()
		if regressor == 'gpr':
			return GaussianProcessRegressor()
		if regressor == 'krr':
			return KernelRidge()
		if regressor == 'knn':
			return KNeighborsRegressor()
		if regressor == 'svr':
			return SVR()
	def fit(self,X,y):
		self._setup_cv()
		if isinstance(self.regressor,str):
			model = self._model_factory(regressor=self.regressor)
			reg_params_cv = self._reformat_grid(self.reg_params)
			GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
			GridSCV.fit(X,y)
			self.best_estimator_ = GridSCV.best_estimator_
			self.best_params_ = GridSCV.best_params_
			self.best_score_ = GridSCV.best_score_
			self.best_overfit_error_ = self.overfit_score(GridSCV)
			self.cv_results_ = GridSCV.cv_results_
		if isinstance(self.regressor,list):
			self.fit_multiple_reg(X,y)
		return self
	def fit_multiple_reg(self,X,y):
		if isinstance(self.regressor,list):
			n_regressors = len(self.regressor)
			self.best_estimators_ = [None]*n_regressors
			self.all_best_params_ = [None]*n_regressors
			self.best_scores_ = np.zeros(n_regressors)
			self.best_overfit_errors_ = [None]*n_regressors
			self.all_cv_results_ = [None]*n_regressors
			assert isinstance(self.reg_params,list), "parameters must also be a list"
			assert len(self.reg_params) == n_regressors, "length of parameters and regressors must match."
			for i,R in enumerate(self.regressor):
				print("Fitting regressor ", R)
				model = self._model_factory(regressor=R)
				reg_params_cv = self._reformat_grid(self.reg_params[i])
				GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
				GridSCV.fit(X,y)
				self.best_estimators_[i] = GridSCV.best_estimator_
				self.all_best_params_[i] = GridSCV.best_params_
				self.best_scores_[i] = GridSCV.best_score_
				self.best_overfit_errors_[i] = self.overfit_score(GridSCV)
				self.all_cv_results_[i] = GridSCV.cv_results_
			self.best_index_ = np.argmax(self.best_scores_)
			self.best_estimator_ = self.best_estimators_[self.best_index_]
			self.best_params_ = self.all_best_params_[self.best_index_]
			self.best_score_ = self.best_scores_[self.best_index_]
			self.best_overfit_error_ = self.best_overfit_errors_[self.best_index_]
			self.cv_results_ = self.all_cv_results_[self.best_index_]
		return self
	def predict(self,X):
		return self.best_estimator_.predict(X)
	def overfit_score(self,GridSCV):
		''' Takes a grid search cv object and returns overfit score '''
		# assert hasattr(self,'GridSCV'), "Must run fit function."
		best_index_ = GridSCV.best_index_
		mean_train_score = GridSCV.cv_results_['mean_train_score']
		mean_test_score = GridSCV.cv_results_['mean_test_score']
		return mean_train_score[best_index_] - mean_test_score[best_index_]

from tqdm import tqdm
class MRegressionWrapperCV(BaseEstimator, RegressorMixin):
	def __init__(self, 
				regressor='pce', reg_params={'order':2},
				target_transform=None,
				target_transform_params={},
				scorer='neg_root_mean_squared_error',
				n_jobs=4,cv=None,
				mixed=True,
				verbose=1):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.cv = cv
		self.target_transform = target_transform
		self.target_transform_params = target_transform_params
		self.mixed = mixed # to control whether mixed surrogate for each target TBD
		self.verbose = verbose
	def _setupCV(self, shuffle=False, randstate=13):
		if self.cv == None:
			self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)
	def _fit_target_transform(self,Y):
		if self.target_transform is None:
			self.TT = FunctionTransformer(lambda Y: Y)
		else:
			if isinstance(self.target_transform,Pipeline):
				print("Target Transform is a pipeline object. Cannot set internal parameters just yet.")
				self.TT = self.target_transform
			else:
				self.TT = self.target_transform(**self.target_transform_params)
		self.TT.fit(Y)
		return self
	def fit(self,X,Y):
		self.n, self.dim = X.shape
		self._setupCV()
		self._fit_target_transform(Y)
		Yhat = self.TT.transform(Y)
		if Yhat.ndim == 1: Yhat = np.atleast_2d(Yhat).T
		assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
		self.ntargets = Yhat.shape[1]
		self._setupCV()
		if isinstance(self.regressor,str):
			self.res = self.fit_single_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == False:
			self.res = self.fit_multiple_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == True:
			self.res = self.fit_single_reg(X,Yhat)
		return self
	def fit_single_reg(self, X, Y, regressor=None,reg_params=None):
		if regressor == None:
			regressor = self.regressor
			reg_params = self.reg_params
		res = defaultdict(list)
		with alive_bar(self.ntargets) as bar:
			for i in range(self.ntargets):
				reg = RegressionWrapperCV(
					regressor=regressor,reg_params=reg_params,
					n_jobs=self.n_jobs, scorer=self.scorer, cv=self.cv, verbose=self.verbose)
				reg.fit(X, Y[:, i])
				res['best_estimators_'].append(reg.best_estimator_)
				res['best_params_'].append(reg.best_params_)
				res['best_scores_'].append(reg.best_score_)
				res['best_overfit_error_'].append(reg.best_overfit_error_)
				res['cv_results_'].append(reg.cv_results_) # cv results from best regressor in list
				res['best_index_'].append(reg.best_index_) # index of best score
				# res['all_cv_results_'].append(reg.all_cv_results_) # all cv results from all regressor in list
				bar()
		self.__dict__.update(res)
		return res
	def fit_multiple_reg(self, X, Y):
		# wont execute if mixed is True
		if isinstance(self.regressor,list):
			mres = []
			for i,r in enumerate(self.regressor):
				res = self.fit_single_reg(X,Y,
								regressor=r,
								reg_params=self.reg_params[i])
				mres.append(res)
		return mres
	def predict(self,X):
		assert hasattr(self,'res'), "Must run fit."
		if isinstance(self.res,dict):
			Ypred = self._predict_single(X,res=self.res)
		elif isinstance(self.res,list):
			Ypred = self._predict_multiple(X,res=self.res)
		return Ypred
	def _predict_single(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,dict), "for single prediction, results must be a dictionary."
		Yhatpred_list = []
		for estimator in self.best_estimators_:
			if X.ndim == 1:
				X = np.atleast_2d(X)
			Yhatpred_list.append(estimator.predict(X))
		Yhatpred = np.array(Yhatpred_list).T
		Ypred = self.TT.inverse_transform(Yhatpred)
		return Ypred
	def _predict_multiple(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,list), "for multiple predictions, results must be a list of dictionaries."
		predictions = []
		for r in res:
			predictions.append(self._predict_single(X,r))
		return np.squeeze(np.array(predictions))
	def feature_importances_(self):
		assert hasattr(self,'res'), "Must run .fit() first!"
		FI_ = []
		for estimator in self.best_estimators_:
			fi = estimator.feature_importances_
			FI_.append(fi)
		return np.array(FI_)
	def score(self,X,Y):
		Ypred = self.predict(X)
		assert Ypred.shape == Y.shape, "predict and Y shape do not match."
		MSPREs = [np.mean((1.0 - Ypred[t]/Y[t])**2) for t in range(Y.shape[0])]
		return np.mean(MSPREs)
	# add def for fitting multiple for each component for faster fitting

class MPCEReg(BaseEstimator, RegressorMixin):
	def __init__(self, 
				regressor='pce', reg_params={'order':2},
				target_transform=None,
				target_transform_params={},
				scorer='neg_root_mean_squared_error',
				n_jobs=4,cv=None,
				mixed=True,
				verbose=1):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.cv = cv
		self.target_transform = target_transform
		self.target_transform_params = target_transform_params
		self.mixed = mixed # to control whether mixed surrogate for each target TBD
		self.verbose = verbose
	def _setupCV(self, shuffle=False, randstate=13):
		if self.cv == None:
			self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)
	def _fit_target_transform(self,Y):
		if self.target_transform is None:
			self.TT = FunctionTransformer(lambda Y: Y)
		else:
			if isinstance(self.target_transform,Pipeline):
				print("Target Transform is a pipeline object. Cannot set internal parameters just yet.")
				self.TT = self.target_transform
			else:
				self.TT = self.target_transform(**self.target_transform_params)
		self.TT.fit(Y)
		return self
	def fit(self,X,Y):
		self.n, self.dim = X.shape
		self._setupCV()
		self._fit_target_transform(Y)
		Yhat = self.TT.transform(Y)
		if Yhat.ndim == 1: Yhat = np.atleast_2d(Yhat).T
		assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
		self.ntargets = Yhat.shape[1]
		self._setupCV()
		if isinstance(self.regressor,str):
			# fits the same regressor to each latent var
			self.res = self.fit_single_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == True:
			# fits different regressor to each latent var
			assert len(self.regressor) == self.ntargets, "number of regressors must be same as the targets."
			assert len(self.reg_params) == self.ntargets, "number of regressors must be same as the targets."
			self.res = self.fit_single_reg(X,Yhat)
		# elif isinstance(self.regressor,list) and self.mixed == True:
		# 	self.res = self.fit_single_reg(X,Yhat)
		return self
	def fit_single_reg(self, X, Y, regressor=None,reg_params=None):
		if regressor == None:
			regressor = self.regressor
			reg_params = self.reg_params
		if isinstance(regressor,str):
			# if single string is provided, repeat it for each target
			regressor = [regressor for i in range(self.ntargets)]
			reg_params = [reg_params for i in range(self.ntargets)]
		res = defaultdict(list)
		with alive_bar(self.ntargets) as bar:
			for i in range(self.ntargets):
				reg = RegressionWrapperCV(
					regressor=[regressor[i]],reg_params=[reg_params[i]],
					n_jobs=self.n_jobs, scorer=self.scorer, cv=self.cv, verbose=self.verbose)
				reg.fit(X, Y[:, i])
				res['best_estimators_'].append(reg.best_estimator_)
				res['best_params_'].append(reg.best_params_)
				res['best_scores_'].append(reg.best_score_)
				res['best_overfit_error_'].append(reg.best_overfit_error_)
				res['cv_results_'].append(reg.cv_results_) # cv results from best regressor in list
				res['best_index_'].append(reg.best_index_) # index of best score
				# res['all_cv_results_'].append(reg.all_cv_results_) # all cv results from all regressor in list
				bar()
		self.__dict__.update(res)
		return res
	def fit_multiple_reg(self, X, Y):
		# wont execute if mixed is True
		if isinstance(self.regressor,list):
			mres = []
			for i,r in enumerate(self.regressor):
				res = self.fit_single_reg(X,Y,
								regressor=r,
								reg_params=self.reg_params[i])
				mres.append(res)
		return mres
	def predict(self,X):
		assert hasattr(self,'res'), "Must run fit."
		if isinstance(self.res,dict):
			Ypred = self._predict_single(X,res=self.res)
		elif isinstance(self.res,list):
			Ypred = self._predict_multiple(X,res=self.res)
		return Ypred
	def _predict_single(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,dict), "for single prediction, results must be a dictionary."
		Yhatpred_list = []
		for estimator in self.best_estimators_:
			if X.ndim == 1:
				X = np.atleast_2d(X)
			Yhatpred_list.append(estimator.predict(X))
		Yhatpred = np.array(Yhatpred_list).T
		Ypred = self.TT.inverse_transform(Yhatpred)
		return Ypred
	def _predict_multiple(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,list), "for multiple predictions, results must be a list of dictionaries."
		predictions = []
		for r in res:
			predictions.append(self._predict_single(X,r))
		return np.squeeze(np.array(predictions))
	def feature_importances_(self):
		assert hasattr(self,'res'), "Must run .fit() first!"
		FI_ = []
		for estimator in self.best_estimators_:
			fi = estimator.feature_importances_
			FI_.append(fi)
		return np.array(FI_)
	def score(self,X,Y):
		Ypred = self.predict(X)
		assert Ypred.shape == Y.shape, "predict and Y shape do not match."
		MSPREs = [np.mean((1.0 - Ypred[t]/Y[t])**2) for t in range(Y.shape[0])]
		return np.mean(MSPREs)
	# add def for fitting multiple for each component for faster fitting


import tesuract
import unittest
import numpy as np
import warnings, pdb
import time as T
import pytest
from sklearn.metrics import mean_squared_error
from sklearn.datasets import make_friedman1
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import KFold
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import make_scorer
from sklearn.model_selection import cross_val_score

from sklearn.decomposition import PCA

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


def mse(a, b):
    return mean_squared_error(a, b, squared=False)


@pytest.mark.regression
class TestPCERegression(unittest.TestCase):
    def test_checking_predict_does_recompute_mindex(self):
        # this test will ensure the model selector selector works
        X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
        p = tesuract.PCEReg(order=2)
        p.fit(X, y)
        p.predict(X)
        Xnew = X[:, :-2]
        p.fit(Xnew, y)
        p.predict(Xnew)
        print("# times computing mindex:", p._mindex_compute_count_)
        assert (
            p._mindex_compute_count_ == 2
        ), "mindex should be recomputed when X dim changes. "

    def test_checking_predict_does_not_recompute_mindex(self):
        # this is a test to count the number of times mindex is computed
        X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
        p = tesuract.PCEReg(order=2)
        p.fit(X, y)
        p.predict(X)
        Xnew, ynew = X[10:], y[10:]
        p.fit(Xnew, ynew)
        p.predict(Xnew)
        print("# times mindex is computed: ", p._mindex_compute_count_)
        assert (
            p._mindex_compute_count_ == 1
        ), "mindex should NOT be recomputed when X row size changes."

    def test_k_fold_mindex_count(self):
        # this is a test to count the number of times mindex is computed
        X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
        X = 2 * X - 1  # scale to [-1,1]
        kf = KFold(n_splits=5)
        p = tesuract.PCEReg()
        for train_index, test_index in kf.split(X):
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]
            p.fit(X_train, y_train)
            score = np.sum(1 - p.predict(X_test) ** 2 / y_test**2)
            # print(score)
        print("# times mindex is computed: ", p._mindex_compute_count_)
        assert (
            p._mindex_compute_count_ == 1
        ), "mindex should NOT be recomputed when X row size changes."

    def test_k_fold_mindex_count(self):
        # this is a test to count the number of times mindex is computed
        X, y = make_friedman1(n_samples=50, n_features=10, random_state=0)
        X = 2 * X - 1  # scale to [-1,1]
        params = {"order": [1, 2, 3]}
        estimator = tesuract.PCEReg(fit_type="LassoCV")
        grid = GridSearchCV(estimator, params)
        grid.fit(X, y)
        # print(grid.cv_results_)
        print("# times computed...", grid.best_estimator_._mindex_compute_count_)
        assert (
            grid.best_estimator_._mindex_compute_count_ == 1
        ), "Best estimator should only compute mindex once!"

    def test_feature_importance_friedman_example(self):
        # this is a test to count the number of times mindex is computed
        X, y = make_friedman1(n_samples=100, n_features=10, random_state=0)
        X = 2 * X - 1  # scale to [-1,1]
        estimator = tesuract.PCEReg(order=4, fit_type="LassoCV")
        estimator.fit(X, y)
        fi = estimator.feature_importances_
        print(fi > 0.05)
        assert (
            np.sum(fi > 0.05) == 5
        ), "friedman function should have 5 important features only."

    def test_grid_search_cv_picks_right_order(self):
        #
        rn = np.random.RandomState(99)
        X = 2 * rn.rand(50, 5) - 1
        fi_index = 3
        x = X[:, 3]
        # y = (1./8)*(35.*x**4 - 30*x**4 + 3)
        y = (1.0 / 2) * (5 * x**3 - 3 * x)
        params = {"order": [1, 2, 3, 4]}
        estimator = tesuract.PCEReg(fit_type="LassoCV")
        start = T.time()
        grid = GridSearchCV(estimator, params, cv=KFold(n_splits=5))
        grid.fit(X, y)
        end = T.time()
        print("Total grid search time is {0:.2f} seconds.".format(end - start))
        # print(grid.cv_results_)
        print("\nbest parameters:", grid.best_params_)
        fi = grid.best_estimator_.feature_importances_
        # print("\nfeatures:", grid.best_estimator_.feature_importances_)
        assert (
            grid.best_params_["order"] == 3
        ), "Grid search should find the best order 4, but it is not. Something is wrong."
        assert (
            np.argmax(fi) == fi_index
        ), "Not finding the right feature importance index."

    def test_grid_search_cv_picks_right_order_and_features(self):
        #
        rn = np.random.RandomState(99)
        X = 2 * rn.rand(50, 5) - 1
        fi_index = 1
        x = X[:, fi_index]
        y = (1.0 / 8) * (35.0 * x**4 - 30 * x**4 + 3)
        # y = (1./2)*(5*x**3 - 3*x)
        params = {"order": [1, 2, 3, 4]}
        estimator = tesuract.PCEReg(fit_type="LassoCV")
        start = T.time()
        grid = GridSearchCV(estimator, params, cv=KFold(n_splits=5))
        grid.fit(X, y)
        end = T.time()
        print("Total grid search time is {0:.2f} seconds.".format(end - start))
        print(grid.cv_results_)
        print("\nbest parameters:", grid.best_params_)
        fi = grid.best_estimator_.feature_importances_
        # print("\nfeatures:", fi)
        assert (
            grid.best_params_["order"] == 3
        ), "Grid search should find the best order 4, but it is not. Something is wrong."
        assert (
            np.argmax(fi) == fi_index
        ), "Not finding the right feature importance index."
        # print("# times computed...", grid.best_estimator_._mindex_compute_count_)


from warnings import simplefilter
from sklearn.exceptions import ConvergenceWarning

simplefilter("ignore", category=ConvergenceWarning)

# regression test for multi output pca regressor
@pytest.mark.regression
class TestMRegressionWrapper(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.X = np.loadtxt(relpath + "/tests/data/rom_test_X.txt")
        self.Y = np.loadtxt(relpath + "/tests/data/rom_test_Y.txt")
        X, Y = self.X, self.Y

        from sklearn.model_selection import KFold

        self.kf = KFold(n_splits=2)

    def test_simplified_model_fit_with_single_param_for_each_comp(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False,
            scale=False,
            pca=True,
            n_components="auto",
            whiten=True,
            exp_var_cutoff=0.5,
        )
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        start = T.time()
        cvscore = cross_val_score(regmodel, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

        n_components = len(regmodel.best_params_)
        reg_custom_list = ["pce" for i in range(n_components)]
        reg_param_list = regmodel.best_params_
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False, scale=False, pca=True, n_components=n_components, whiten=True
        )
        regmodel_opt = tesuract.MRegressionWrapperCV(
            regressor=reg_custom_list,
            reg_params=reg_param_list,
            custom_params=True,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        start = T.time()
        cvscore = cross_val_score(regmodel_opt, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

    def test_simplified_model_fit_with_uninstantiated_PCA_target_transform(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        # target_transform = tesuract.preprocessing.target_pipeline_custom(log=False,scale=False,pca=True,n_components='auto',whiten=True,cutoff=.5)
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=PCA,
            target_transform_params={"n_components": 4},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel._fit_target_transform(Y)

    def test_target_transform_has_certain_attributes_without_target_instantiation(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        # FactorAnalysis which has no inverse transform
        from sklearn.decomposition import FactorAnalysis

        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=FactorAnalysis,
            target_transform_params={"n_components": 8},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        with self.assertRaises(Exception):
            regmodel._fit_target_transform(Y)

    def test_target_transform_has_certain_attributes_w_target_instantiation(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        # FactorAnalysis which has no inverse transform
        from sklearn.decomposition import FactorAnalysis

        FA = FactorAnalysis(n_components=4)
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=FA,
            target_transform_params=None,
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        with self.assertRaises(Exception):
            regmodel._fit_target_transform(Y)

    def test_simplified_model_fit_with_instantiated_PCA_target_transform(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        # target_transform = tesuract.preprocessing.target_pipeline_custom(log=False,scale=False,pca=True,n_components='auto',whiten=True,cutoff=.5)
        target_transform = PCA(n_components=2)
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        start = T.time()
        cvscore = cross_val_score(regmodel, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

        n_components = len(regmodel.best_params_)
        reg_custom_list = ["pce" for i in range(n_components)]
        reg_param_list = regmodel.best_params_
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False, scale=False, pca=True, n_components=n_components, whiten=True
        )
        regmodel_opt = tesuract.MRegressionWrapperCV(
            regressor=reg_custom_list,
            reg_params=reg_param_list,
            custom_params=True,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        start = T.time()
        cvscore = cross_val_score(regmodel_opt, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

    def test_multi_target_init_with_custom_param_list(self):
        # uses the best params as the new set of param grid space
        # not efficient but can be faster than original space
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False,
            scale=False,
            pca=True,
            n_components="auto",
            whiten=True,
            exp_var_cutoff=0.5,
        )
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        start = T.time()
        cvscore = cross_val_score(regmodel, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

        new_params = regmodel.best_params_
        new_regressors = ["pce" for i in range(len(new_params))]
        regmodel_opt = tesuract.MRegressionWrapperCV(
            regressor=new_regressors,
            reg_params=new_params,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        start = T.time()
        cvscore = cross_val_score(regmodel_opt, X, Y, cv=5, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        print("total time is ", T.time() - start)

    def test_multi_target_init_with_custom_param_list2(self):
        # uses the best params as the new set of param grid space
        # not efficient but can be faster than original space
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = "pce"  # will be a list automaticall if string
        param_list = pce_grid

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.PCATargetTransform(
            n_components="auto",
            whiten=True,
            exp_var_cutoff=0.5,
        )
        print(target_transform.fit_transform(Y).shape)
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        print("custom params:", regmodel.custom_params)  # should be
        regmodel.fit(X, Y)

    def test_rom_w_single_regressor_as_list(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False, scale=False, pca=True, n_components=4, whiten=True
        )
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        cvscore = cross_val_score(regmodel, X, Y, cv=2, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))

    def test_rom_w_single_regressor_as_list_wo_pca(self):
        X, Y = self.X, self.Y[:, ::75]  # shorten output
        pce_grid = [
            {
                "order": list(range(2)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = ["pce"]
        param_list = [pce_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        # target_transform = tesuract.preprocessing.target_pipeline_custom(log=False,scale=False,pca=False,n_components=4,whiten=True)
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=None,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        cvscore = cross_val_score(regmodel, X, Y, cv=2, scoring="r2", n_jobs=1)
        print("cv r2 score: {0}%".format(-100 * np.round(cvscore.mean(), 4)))
        # get feature importances and weighted versions
        fi = regmodel.feature_importances_
        ws = regmodel.sobol_weighted_
        assert ws.ndim == 1, "weighted sobol must be a single vector"
        # get explained variance ratio
        evr = regmodel.explained_variance_ratio_
        assert np.amin(evr) == np.amax(
            evr
        ), "The explained variance ratio when target transform = None should be equal."
        # print("explained var ratio when target transform is None: ", evr)

    def test_rom_w_single_regressor_as_str(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        regressors = "pce"
        param_list = pce_grid

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False,
            scale=False,
            pca=True,
            n_components=3,
            whiten=True,
            exp_var_cutoff=1 - 0.25,
        )
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)
        # test feature importance and explained var ratio for pca
        fi = regmodel.feature_importances_
        ws = regmodel.sobol_weighted_
        assert ws.ndim == 1, "weighted sobol must be a single vector"
        # get explained variance ratio
        evr = regmodel.explained_variance_ratio_
        print("explained variance ratio when using pca:", evr)

    def test_rom_w_multiple_regressors(self):
        X, Y = self.X, self.Y
        pce_grid = [
            {
                "order": list(range(1, 3)),
                "mindex_type": ["total_order"],
                "fit_type": ["LassoCV"],
            }
        ]
        # random forest fit
        rf_grid = {
            "n_estimators": [50, 100],
            "max_features": ["auto"],
            "max_depth": [5, 10],
        }
        regressors = ["pce", "rf"]
        param_list = [pce_grid, rf_grid]

        def my_scorer(ytrue, ypred):
            mse = np.mean((ytrue - ypred) ** 2) / np.mean(ytrue**2)
            return -mse

        custom_scorer = make_scorer(my_scorer, greater_is_better=True)
        target_transform = tesuract.preprocessing.target_pipeline_custom(
            log=False,
            scale=False,
            pca=True,
            n_components="auto",
            whiten=True,
            exp_var_cutoff=1 - 0.25,
        )
        regmodel = tesuract.MRegressionWrapperCV(
            regressor=regressors,
            reg_params=param_list,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=1,
            scorer=custom_scorer,
            verbose=0,
        )
        regmodel.fit(X, Y)



import tesuract
import unittest
import numpy as np
import pytest
from tesuract.preprocessing import *
from sklearn.datasets import make_low_rank_matrix

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


@pytest.mark.unit
class Test_Target_Scaler(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rn = np.random.RandomState(399123)

    def test_target_scaler(self):
        Y = 3.5 * self.rn.rand(10, 100) + 2
        scaler = MinMaxTargetScaler()
        scaler.fit(Y)
        Yhat = scaler.fit_transform(Y)
        assert (
            np.amin(Yhat) == 0.0 and np.amax(Yhat) == 1.0
        ), "Scaler is not working properly"

    def test_target_scaler_inverse(self):
        Y = 3.5 * self.rn.rand(10, 100) + 2
        scaler = MinMaxTargetScaler()
        scaler.fit(Y)
        Yhat = scaler.fit_transform(Y)
        assert (
            np.sum(scaler.inverse_transform(Yhat) - Y) <= 1e-10
        ), "Inverse transform not working properly."

    def test_minmaxtargetscaler(self):
        Y = self.rn.rand(10, 3)
        ts = MinMaxTargetScaler(target_range=(1.0, 2.0))
        ts.fit(Y)
        Yt = ts.transform(Y)
        assert np.amin(Yt) >= 1.0, "transform out of range"
        assert np.amax(Yt) <= 2.0, "transform out of range"
        error = np.linalg.norm(ts.inverse_transform(Yt) - Y)
        assert error <= 1e-15, "transform and inverse not working"

    def test_simple_mapping(self):
        Y = np.array([np.zeros(3), np.ones(3)])
        ts = MinMaxTargetScaler(target_range=(3.0, 10.0))
        ts.fit(Y)
        Yt = ts.transform(Y)
        assert np.mean(Yt[0]) == 3.0, "transform out of range"
        assert np.mean(Yt[1]) == 10.0, "transform out of range"

    def test_simple_mapping2(self):
        Y = np.array([3 + np.zeros(3), 5 * np.ones(3)])
        ts = MinMaxTargetScaler(target_range=(0.0, 1.0))
        ts.fit(Y)
        Yt = ts.transform(Y)
        assert np.mean(Yt[0]) == 0.0, "transform out of range"
        assert np.mean(Yt[1]) == 1.0, "transform out of range"


@pytest.mark.unit
class Test_Domain_Scaler(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rn = np.random.RandomState(399123)

    def test_domain_scaler(self):
        ds = DomainScaler(dim=3, input_range=[(1.0, 3.0), (1.0, 3.0), (1.0, 3.0)])
        Y = np.array([1 + np.zeros(3), 3 * np.ones(3)])
        Yt = ds.fit_transform(Y)
        assert np.mean(Yt[0]) == -1.0, "transform is broken"
        assert np.mean(Yt[1]) == 1.0, "transform is broken"
        error = np.linalg.norm(ds.inverse_transform(Yt) - Y)
        assert error <= 1e-15, "inverse transform not working"

    def test_nonuniform_input_range(self):
        ds = DomainScaler(dim=3, input_range=[(1.0, 3.0), (1.0, 4.0), (1.0, 5.0)])
        Y = np.array([[1, 1, 1], [3.0, 4.0, 5.0]])
        Yt = ds.fit_transform(Y)
        assert np.mean(Yt[0]) == -1.0, "transform is broken"
        assert np.mean(Yt[1]) == 1.0, "transform is broken"
        error = np.linalg.norm(ds.inverse_transform(Yt) - Y)
        assert error <= 1e-15, "inverse transform not working"

    def test_nonuniform_input_range2(self):
        ds = DomainScaler(dim=3, input_range=[(1.0, 3.0), (1.1, 4.0), (1.2, 5.0)])
        Y = np.array([[1, 1.1, 1.2], [3.0, 4.0, 5.0]])
        Yt = ds.fit_transform(Y)
        assert np.mean(Yt[0]) == -1.0, "transform is broken"
        assert np.mean(Yt[1]) == 1.0, "transform is broken"
        error = np.linalg.norm(ds.inverse_transform(Yt) - Y)
        assert error <= 1e-15, "inverse transform not working"

    def test_nondefault_input_output_range(self):
        ds = DomainScaler(
            dim=3,
            input_range=[(1.0, 3.0), (1.1, 4.0), (1.2, 5.0)],
            output_range=(10, 11.0),
        )
        Y = np.array([[1, 1.1, 1.2], [3.0, 4.0, 5.0]])
        Yt = ds.fit_transform(Y)
        assert np.mean(Yt[0]) == 10.0, "transform is broken"
        assert np.mean(Yt[1]) == 11.0, "transform is broken"
        error = np.linalg.norm(ds.inverse_transform(Yt) - Y)
        assert error <= 1e-15, "inverse transform not working"


@pytest.mark.unit
class Test_Log_Scaler(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rn = np.random.RandomState(399123)
        self.Y = self.rn.rand(100, 5)
        self.LT = LogTransform()

    def test_log_transform(self):
        Y = self.Y
        logY = np.log(Y)
        self.LT.fit(Y)
        assert (
            np.linalg.norm(self.LT.transform(Y) - logY) < 1e-15
        ), ".transform method is not working. "

    def test_log_inverse_transform(self):
        Y = self.Y
        logY = np.log(Y)
        self.LT.fit(Y)
        Yhat = self.LT.transform(Y)
        assert (
            np.linalg.norm(self.LT.inverse_transform(Yhat) - Y) < 1e-15
        ), ".transform method is not working. "

    def test_log_transform_neg_check(self):
        Y = self.rn.randn(100, 5)
        with pytest.raises(AssertionError):
            self.LT.transform(Y)


@pytest.mark.unit
class Test_Pipeline(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rn = np.random.RandomState(399123)
        self.Y = self.rn.rand(10, 5)
        self.LT = LogTransform()

    def test_gen_pipeline_log(self):
        ncomp = 4
        pipe = gen_scale_pipeline(log=True, scale=False, pca=False, n_components=ncomp)
        Y = self.Y
        pipe.fit(Y)
        Yhat = pipe.transform(Y)
        Yr = pipe.inverse_transform(Yhat)
        # assert Yhat.shape[1] == ncomp, "new dimensions do not match n_components. Check PCA."
        assert Yr.shape == Y.shape, "original and reconstructed shapes do no match. "
        assert np.mean(np.mean((Yr - Y) ** 2, axis=1)) < 1e-16

    def test_gen_pipeline_log_scale(self):
        ncomp = 4
        pipe = gen_scale_pipeline(log=True, scale=True, pca=False, n_components=ncomp)
        Y = self.Y
        pipe.fit(Y)
        Yhat = pipe.transform(Y)
        Yr = pipe.inverse_transform(Yhat)
        # assert Yhat.shape[1] == ncomp, "new dimensions do not match n_components. Check PCA."
        assert Yr.shape == Y.shape, "original and reconstructed shapes do no match. "
        assert np.mean(np.mean((Yr - Y) ** 2, axis=1)) < 1e-16

    def test_gen_pipeline_log_scale_pca(self):
        ncomp = self.Y.shape[1]
        pipe = gen_scale_pipeline(
            log=True, scale=True, pca=True, n_components=ncomp, svd_solver="auto"
        )
        Y = self.Y
        pipe.fit(Y)
        Yhat = pipe.transform(Y)
        Yr = pipe.inverse_transform(Yhat)
        assert (
            Yhat.shape[1] == ncomp
        ), "new dimensions do not match n_components. Check PCA."
        assert Yr.shape == Y.shape, "original and reconstructed shapes do no match. "
        rmse = np.mean((Yr - Y) ** 2, axis=1) / np.mean(Y**2, axis=1)
        assert (
            rmse.mean() < 1e-16
        ), "PCA should be exact, but isn't. Something is wrong."


@pytest.mark.unit
class Test_PCA_Target_Transform(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rn = np.random.RandomState(399123)
        self.K = 8
        self.Y = make_low_rank_matrix(
            n_samples=100,
            n_features=75,
            effective_rank=self.K,
            tail_strength=0.01,
            random_state=self.rn,
        )
        U, S, VT = np.linalg.svd(self.Y, full_matrices=False)
        S[: self.K] = 0.0
        self.Y = np.dot(np.dot(U, np.diag(S)), VT)

    def test_pca_auto(self):
        Y = self.Y
        # pca1 = PCA(n_components=50) # sklearn's PCA
        cutoff = 1e-2
        pca = PCATargetTransform(n_components="auto", exp_var_cutoff=1 - cutoff)
        pca.fit(Y)
        assert 1 - pca.cumulative_error[-1] <= cutoff, "cutoff not working. "
        assert pca.K <= self.K + 2, "auto cutoff isn't working to desired amount."

    def test_pca_cumulative_error(self):
        Y = self.Y
        pca1 = PCA(n_components=50)  # sklearn's PCA
        pca2 = PCATargetTransform(n_components=20)
        pca1.fit(Y)
        pca2.fit(Y)
        ce1 = np.cumsum(pca1.explained_variance_ratio_)[: pca2.K]
        ce2 = pca2.cumulative_error
        assert np.linalg.norm(ce1 - ce2) / np.linalg.norm(ce1) <= 1e-12



import numpy as np 
import tesuract
from sklearn.datasets import make_friedman1
import time as T

# generate test data
rn = np.random.RandomState(44)
dim = 10
X,y = make_friedman1(n_samples=100,n_features=dim,random_state=rn)


# transform data to [-1,1] for polynomials
dScaler = tesuract.preprocessing.DomainScaler(dim=dim,input_range=(0,1))
Xs = dScaler.fit_transform(X)

# fit using PCE CV wrapper
pce_param_grid = \
	{'order': [1,2,3,4],
	 'mindex_type': ['total_order', 'hyperbolic', ],
	 'fit_type': ['LassoCV', 'linear', 'ElasticNetCV']}

rf_param_grid = \
	{'n_estimators': [100,300,750],
	 'max_features': ['sqrt','auto'],
	 'max_depth': [5,10,25]}

rf_param_grid = \
	{'n_estimators': [100,300,750],
	 'max_features': ['sqrt','auto'],
	 'max_depth': [5,10,25]}

svr_param_grid = {
	'kernel': ('linear','poly', 'rbf', 'sigmoid'),
	'degree': (2,4,8),
	'gamma': ('scale','auto'),
	'C': (1,5,10)
} 

# start = T.time()
# pceCV = tesuract.RegressionWrapperCV(regressor='pce',reg_params=pce_param_grid,n_jobs=-1)
# pceCV.fit(Xs,y)
# end = T.time()
# print("best pce error: {0:.5f}, {1:.5f} sec".format(-pceCV.best_score_,end-start))

# PCA test
dim = 250
samples = 500
mean = np.zeros(dim)
t = np.linspace(0,1,dim)
tx,ty = np.meshgrid(t,t)
C = np.array([min(pair) for pair in zip(tx.reshape(-1),ty.reshape(-1))]).reshape(dim,dim)
Y = rn.multivariate_normal(mean,C,size=(samples,))

from sklearn.decomposition import PCA
# pca = PCA(n_components='mle',whiten=True)
pca = tesuract.preprocessing.PCATargetTransform(n_components='auto',cutoff=.05,whiten=False)
Yhat = pca.fit_transform(Y)
components = pca.components_
print(pca.n_components_)

import matplotlib.pyplot as plt
fig, ax = plt.subplots(1)
ysamples = [yhat for yhat in Yhat.T]
# for yi in ysamples: ax.hist(yi,bins=30)

# # try with repeated k fold cv
# from sklearn.model_selection import RepeatedKFold
# rkcv = RepeatedKFold(n_splits=5,n_repeats=10)
# start = T.time()
# pceCV2 = tesuract.RegressionWrapperCV(regressor='pce',reg_params=pce_param_grid,cv=rkcv,n_jobs=-1)
# pceCV2.fit(Xs,y)
# end = T.time()
# print("best pce error: {0:.5f}, {1:.5f} sec".format(-pceCV2.best_score_,end-start))

# rfCV = tesuract.RegressionWrapperCV(regressor='rf',reg_params=rf_param_grid,n_jobs=-1)
# rfCV.fit(Xs,y)
# print("best rf score:", rfCV.best_score_)

# svrCV = tesuract.RegressionWrapperCV(regressor='svr',reg_params=svr_param_grid,n_jobs=-1)
# svrCV.fit(Xs,y)
# print("best svr score:", svrCV.best_score_)

# multimodelCV = tesuract.RegressionWrapperCV(regressor=['pce','rf','svr'],reg_params=[pce_param_grid,rf_param_grid,svr_param_grid],n_jobs=-1)
# multimodelCV.fit(Xs,y)
# print("best pce/rf score:", multimodelCV.best_score_)




import tesuract
import unittest
import numpy as np
from tesuract.preprocessing import DomainScaler, MinMaxTargetScaler
import pytest

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


@pytest.mark.unit
class TestTotalOrderMultiIndex(unittest.TestCase):
    def setUp(self):
        self.dim = 3
        self.order = 5
        self.M = tesuract.MultiIndex(dim=self.dim, order=self.order)
        self.mtest = np.loadtxt(relpath + "/tests/data/multindex_d3_o5.txt")

    def test_multindex(self):
        abserr = np.sum(self.M.index - self.mtest)
        assert abserr == 0, "multindex for d=3,o=5 is not right."

    def test_verify_max_order(self):
        self.assertTrue(np.all(np.sum(self.M.index, 1) <= self.order))

    def test_multiindex_length_and_dim(self):
        factorial = np.math.factorial
        nBasis = (
            factorial(self.order + self.dim)
            / factorial(self.order)
            / factorial(self.dim)
        )
        self.assertEqual(self.M.index.shape[0], nBasis)
        self.assertEqual(self.M.index.shape[1], self.dim)

    def test_nPCTerms(self):
        self.assertEqual(self.M.index.shape[0], self.M.nPCTerms)

    def test_set_index(self):
        nPCTerms_old = self.M.nPCTerms
        test_index = self.M.index[:-1]
        self.M.setIndex(test_index)
        self.assertEqual(self.M.nPCTerms, nPCTerms_old - 1)


@pytest.mark.unit
class TestHyperbolicMultiIndex(unittest.TestCase):
    def setUp(self):
        self.dim = 3
        self.order = 4
        self.M = tesuract.MultiIndex(
            dim=self.dim, order=self.order, mindex_type="hyperbolic"
        )
        self.mtest = np.loadtxt(relpath + "/tests/data/hypmultindex_d3_o4.txt")

    def test_hyperbolic_multindex(self):
        abserr = np.sum(self.M.index - self.mtest)
        assert abserr == 0, "hyperbolic multindex for d=3,o=4 is not right."

    def test_verify_max_order(self):
        self.assertTrue(np.all(np.sum(self.M.index, 1) <= self.order))

    def test_multiindex_length_and_dim(self):
        self.assertEqual(self.M.index.shape[0], 16)
        self.assertEqual(self.M.index.shape[1], self.dim)

    def test_nPCTerms(self):
        self.assertEqual(self.M.index.shape[0], self.M.nPCTerms)



import tesuract
import unittest
import numpy as np
import warnings, pdb
import time as T

from sklearn.model_selection import RepeatedKFold
import pytest

relpath = tesuract.__file__[:-11]  # ignore the __init__.py specification
print(relpath)


def mse(a, b):
    return mean_squared_error(a, b, squared=False)


import numpy as np
import tesuract
from sklearn.datasets import make_friedman1

# generate test data
rn = np.random.RandomState(44)
dim = 10
X, y = make_friedman1(n_samples=100, n_features=dim, random_state=rn)

# transform data to [-1,1] for polynomials
dScaler = tesuract.preprocessing.DomainScaler(dim=dim, input_range=(0, 1))
Xs = dScaler.fit_transform(X)

# fit using PCE CV wrapper
pce_param_grid = {
    "order": [1],
    "mindex_type": ["total_order"],
    "fit_type": ["LassoCV", "linear"],
}

rf_param_grid = {"n_estimators": [100], "max_features": ["auto"], "max_depth": [5, 10]}

svr_param_grid = {
    "kernel": ["linear", "poly"],
    "degree": [4],
    "gamma": ["auto"],
    "C": [1],
}


@pytest.mark.regression
class TestPCERegression(unittest.TestCase):
    def test_pce_cv_regression(self):
        pceCV = tesuract.RegressionWrapperCV(
            regressor="pce", reg_params=pce_param_grid, n_jobs=-1
        )
        pceCV.fit(Xs, y)
        print("best pce score:", pceCV.best_score_)
        print(
            "best pce faeture importances:", pceCV.best_estimator_.feature_importances_
        )

    def test_pce_repeated_kfold_cv_regression(self):
        rkcv = RepeatedKFold(n_splits=5, n_repeats=2)
        pceCV2 = tesuract.RegressionWrapperCV(
            regressor="pce", reg_params=pce_param_grid, cv=rkcv, n_jobs=-1
        )
        pceCV2.fit(Xs, y)
        print("best pce error: {0:.5f}".format(-pceCV2.best_score_))
        print(
            "best pce faeture importances:", pceCV2.best_estimator_.feature_importances_
        )

    def test_rf_cv_regression(self):
        rfCV = tesuract.RegressionWrapperCV(
            regressor="rf", reg_params=rf_param_grid, n_jobs=-1
        )
        rfCV.fit(Xs, y)
        print("best rf score:", rfCV.best_score_)

    def test_svr_cv_regression(self):
        svrCV = tesuract.RegressionWrapperCV(
            regressor="svr", reg_params=svr_param_grid, n_jobs=-1
        )
        svrCV.fit(Xs, y)
        print("best svr score:", svrCV.best_score_)

    def test_multimodel_cv_regression(self):
        multimodelCV = tesuract.RegressionWrapperCV(
            regressor=["pce", "rf", "svr"],
            reg_params=[pce_param_grid, rf_param_grid, svr_param_grid],
            n_jobs=-1,
        )
        multimodelCV.fit(Xs, y)
        print("best pce/rf score:", multimodelCV.best_score_)



import jax
import jax.numpy as jnp
import numpy as onp
import scipy

# This needs to run at startup
# https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#Double-(64bit)-precision
jax.config.update('jax_enable_x64', True)

def run(np,optname):
    print(f"\nRunning {optname} on {np}")
    def rosen(x):
        return np.sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)

    # double precision is needed for lbfgsb and probably other optimisers
    # https://github.com/google/jax/issues/936
    # https://github.com/scipy/scipy/issues/5832
    x0 = np.array([1.3, 0.7, 0.8, 1.9, 1.2], dtype='float64')

    optopt = {'disp': True}

    if np==onp:
      return scipy.optimize.minimize(rosen, x0, method=optname, options=optopt)
    else:
      g_rosen = jax.grad(rosen)

      g_rosen_used = g_rosen
      if optname=='L-BFGS-B':
        # Need to make sure the data is copy from jax for LBFGSB
        # https://github.com/google/jax/issues/1510
        # asarray is not sufficient
        # g_rosen_as_np = lambda x:onp.asarray(jax.jit(g_rosen)(x))
        g_rosen_np = lambda x:onp.array(jax.jit(g_rosen)(x))
        g_rosen_used = g_rosen_np

      return scipy.optimize.minimize(rosen, x0, jac=g_rosen_used, method=optname, options=optopt)



import numpy as np
import tesuract
from collections import defaultdict
import time as T
import warnings, pdb

from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.estimator_checks import check_estimator
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import (
    GridSearchCV,
    ParameterGrid,
    KFold,
    cross_val_score,
    cross_validate,
)
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import FunctionTransformer
from sklearn.metrics import make_scorer
from sklearn.decomposition import PCA, FastICA
from sklearn.pipeline import Pipeline
from sklearn.neural_network import MLPRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR

# from alive_progress import alive_bar
from tqdm import tqdm

from warnings import simplefilter
from sklearn.exceptions import ConvergenceWarning


class RegressionWrapperCV(BaseEstimator):
    def __init__(
        self,
        regressor="pce",
        reg_params={},
        scorer="neg_root_mean_squared_error",
        n_jobs=1,
        verbose=0,
        cv=None,
    ):
        self.regressor = regressor
        self.reg_params = reg_params
        self.scorer = scorer
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.cv = cv
        # self.__dict__.update(reg_params)

    def _setup_cv(self):
        if self.cv == None:
            self.cv = KFold(n_splits=5)

    def _reformat_grid(self, params):
        if isinstance(self.regressor, str):
            self.regressor = [self.regressor]
        try:
            # if parameter grid is in the form of grid search cv
            ParameterGrid(params)
            params_cv = params
        except:
            # convert reg_params to form that can be used by grid search
            # for a single grid point
            params_cv = defaultdict(list)
            for key, item in params.items():
                params_cv[key].append(item)
        return params_cv

    def _model_factory(self, regressor):
        if regressor == "pce":
            return tesuract.PCEReg()
        if regressor == "rf":
            return RandomForestRegressor()
        if regressor == "mlp":
            return MLPRegressor()
        if regressor == "gpr":
            return GaussianProcessRegressor()
        if regressor == "krr":
            return KernelRidge()
        if regressor == "knn":
            return KNeighborsRegressor()
        if regressor == "svr":
            return SVR()

    def fit(self, X, y):
        self._setup_cv()
        if isinstance(self.regressor, str):
            self.regressor = [self.regressor]
        self.fit_multiple_reg(X, y)
        return self

    def fit_multiple_reg(self, X, y):
        n_regressors = len(self.regressor)
        self.best_estimators_ = [None] * n_regressors
        self.all_best_params_ = [None] * n_regressors
        self.best_scores_ = np.zeros(n_regressors)
        self.best_overfit_errors_ = [None] * n_regressors
        self.all_cv_results_ = [None] * n_regressors
        # potentially depricated
        # assert isinstance(self.reg_params,list), "parameters must also be a list"
        if isinstance(self.reg_params, list) is False:
            self.reg_params = [self.reg_params]
        # Note: length of regressors does not need to match length of grid params since grid params can be a list of different combinations
        # assert (
        #     len(self.reg_params) == n_regressors
        # ), "length of parameters and regressors must match."
        for i, R in enumerate(self.regressor):
            # print("Fitting regressor ", R)
            model = self._model_factory(regressor=R)
            reg_params_cv = self._reformat_grid(self.reg_params[i])
            GridSCV = GridSearchCV(
                model,
                reg_params_cv,
                scoring=self.scorer,
                n_jobs=self.n_jobs,
                cv=self.cv,
                verbose=self.verbose,
                return_train_score=True,
            )
            GridSCV.fit(X, y)
            self.best_estimators_[i] = GridSCV.best_estimator_
            self.all_best_params_[i] = GridSCV.best_params_
            self.best_scores_[i] = GridSCV.best_score_
            self.best_overfit_errors_[i] = self.overfit_score(GridSCV)
            self.all_cv_results_[i] = GridSCV.cv_results_
        self.best_index_ = np.argmax(self.best_scores_)
        self.best_estimator_ = self.best_estimators_[self.best_index_]
        self.best_params_ = self.all_best_params_[self.best_index_]
        self.best_score_ = self.best_scores_[self.best_index_]
        self.best_overfit_error_ = self.best_overfit_errors_[self.best_index_]
        self.cv_results_ = self.all_cv_results_[self.best_index_]
        return self

    def predict(self, X):
        return self.best_estimator_.predict(X)

    def overfit_score(self, GridSCV):
        """Takes a grid search cv object and returns overfit score"""
        # assert hasattr(self,'GridSCV'), "Must run fit function."
        best_index_ = GridSCV.best_index_
        mean_train_score = GridSCV.cv_results_["mean_train_score"]
        mean_test_score = GridSCV.cv_results_["mean_test_score"]
        return mean_train_score[best_index_] - mean_test_score[best_index_]


class MRegressionWrapperCV(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        regressor="pce",
        reg_params={"order": 2},
        custom_params=False,
        target_transform=None,
        target_transform_params={},
        scorer="neg_root_mean_squared_error",
        n_jobs=-1,
        cv=None,
        mixed=True,
        verbose=0,
    ):
        self.regressor = regressor
        self.reg_params = reg_params
        self.custom_params = custom_params
        self.scorer = scorer
        self.n_jobs = n_jobs
        self.cv = cv
        self.target_transform = target_transform
        self.target_transform_params = target_transform_params
        self.mixed = mixed  # to control whether mixed surrogate for each target TBD
        self.verbose = verbose

    def _setupCV(self, shuffle=False, randstate=13):
        if self.cv == None:
            self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)

    def _fit_target_transform(self, Y):
        if self.target_transform is None:
            self.TT = FunctionTransformer(lambda Y: Y)
        else:
            if isinstance(self.target_transform, Pipeline):
                # print("Target Transform is a pipeline object. Cannot set internal parameters just yet.")
                self.TT = self.target_transform
            # elif isinstance(self.target_transform,PCA):
            # 	# if instantiated PCA object, ok to set to target transform
            # 	self.TT = self.target_transform
            else:
                if isinstance(self.target_transform, type):
                    # Allow user to pass uninstantiated class
                    # for uninstantiated input to target transform
                    self.TT = self.target_transform(**self.target_transform_params)
                else:
                    # Allow user to pass an instantiated class
                    self.TT = self.target_transform
                assert hasattr(
                    self.target_transform, "fit"
                ), "transform object is instantiated and must have a fit method."
                assert hasattr(
                    self.target_transform, "transform"
                ), "transform object is instantiated and must have a transform method."
                assert hasattr(
                    self.target_transform, "inverse_transform"
                ), "transform object is instantiated and must have an inverse_transform method."

        self.TT.fit(Y)
        return self

    def fit(self, X, Y):
        self.n, self.dim = X.shape
        self._setupCV()
        self._fit_target_transform(Y)
        Yhat = self.TT.transform(Y)
        # save explained variance ratio if using PCA
        # otherwise, return equal weights
        try:
            self.explained_variance_ratio_ = (
                self.TT.named_steps.pca.explained_variance_ratio_
            )
            self.cumsum_explained_variance_ratio_ = (
                self.TT.named_steps.pca.cumulative_error_full
            )
        except:
            self.explained_variance_ratio_ = np.ones(Yhat.shape[1]) / Yhat.shape[1]
        # compute total variance for feature importances
        self.TotVar = np.var(
            Y
        )  # .var(axis=0).mean()# total variance averaged over all space
        if Yhat.ndim == 1:
            Yhat = np.atleast_2d(Yhat).T
        assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
        self.ntargets = Yhat.shape[1]
        self._setupCV()
        if isinstance(self.regressor, str):
            self.res = self.fit_single_reg(X, Yhat)
        elif isinstance(self.regressor, list) and self.mixed == False:
            self.res = self.fit_multiple_reg(X, Yhat)
        elif isinstance(self.regressor, list) and self.mixed == True:
            self.res = self.fit_single_reg(X, Yhat)
        return self

    def fit_single_reg(self, X, Y, regressor=None, reg_params=None):
        if regressor == None:
            regressor = self.regressor
            if isinstance(regressor, str):
                regressor = [regressor]
            reg_params = self.reg_params
        # elif len(reg_params) == 1:
        # 	reg_params = [[reg_params] for i in range(self.ntargets)]
        # if isinstance(regressor,list):
        # 	# if regressor is a list and mixed is True, repeat regressor params
        # 	# this is to keep consistent the case when we custom input regressors for each target
        # 	reg_params = [reg_params for i in range(len(regressor))]
        # 	print(reg_params, regressor)
        res = defaultdict(list)
        # with alive_bar(self.ntargets) as bar:
        for i in tqdm(range(self.ntargets)):
            if self.custom_params:
                # fit a single regressor to each target
                assert len(regressor) == len(reg_params)
                assert len(regressor) == self.ntargets
                reg = RegressionWrapperCV(
                    regressor=[regressor[i]],
                    reg_params=[reg_params[i]],
                    n_jobs=self.n_jobs,
                    scorer=self.scorer,
                    cv=self.cv,
                    verbose=self.verbose,
                )
                reg.fit(X, Y[:, i])
                res["best_estimators_"].append(reg.best_estimator_)
                res["best_params_"].append(reg.best_params_)
                res["best_scores_"].append(reg.best_score_)
                dict_scores_ = {regressor[i]: reg.best_scores_}
                res["best_scores_all_"].append(dict_scores_)
                res["best_overfit_error_"].append(reg.best_overfit_error_)
                res["cv_results_"].append(
                    reg.cv_results_
                )  # cv results from best regressor in list
                res["best_index_"].append(reg.best_index_)  # index of best score
                # res['all_cv_results_'].append(reg.all_cv_results_) # all cv results from all regressor in list
            else:
                reg = RegressionWrapperCV(
                    regressor=regressor,
                    reg_params=reg_params,
                    n_jobs=self.n_jobs,
                    scorer=self.scorer,
                    cv=self.cv,
                    verbose=self.verbose,
                )
                reg.fit(X, Y[:, i])
                res["best_estimators_"].append(reg.best_estimator_)
                res["best_params_"].append(reg.best_params_)
                res["best_scores_"].append(reg.best_score_)
                dict_scores_ = {
                    regressor[j]: reg.best_scores_[j] for j in range(len(regressor))
                }
                res["best_scores_all_"].append(dict_scores_)
                res["best_overfit_error_"].append(reg.best_overfit_error_)
                res["cv_results_"].append(
                    reg.cv_results_
                )  # cv results from best regressor in list
                res["best_index_"].append(reg.best_index_)  # index of best score
                # res['all_cv_results_'].append(reg.all_cv_results_) # all cv results from all regressor in list
            # bar()
        self.__dict__.update(res)
        return res

    def fit_multiple_reg(self, X, Y):
        # wont execute if mixed is True
        if isinstance(self.regressor, list):
            mres = []
            for i, r in enumerate(self.regressor):
                res = self.fit_single_reg(
                    X, Y, regressor=r, reg_params=self.reg_params[i]
                )
                mres.append(res)
        return mres

    def predict(self, X):
        assert hasattr(self, "res"), "Must run fit."
        if isinstance(self.res, dict):
            Ypred = self._predict_single(X, res=self.res)
        elif isinstance(self.res, list):
            Ypred = self._predict_multiple(X, res=self.res)
        return Ypred

    def _predict_single(self, X, res=None):
        # assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
        assert isinstance(
            res, dict
        ), "for single prediction, results must be a dictionary."
        Yhatpred_list = []
        for estimator in self.best_estimators_:
            if X.ndim == 1:
                X = np.atleast_2d(X)
            Yhatpred_list.append(estimator.predict(X))
        Yhatpred = np.array(Yhatpred_list).T
        Ypred = self.TT.inverse_transform(Yhatpred)
        return Ypred

    def _predict_multiple(self, X, res=None):
        # assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
        assert isinstance(
            res, list
        ), "for multiple predictions, results must be a list of dictionaries."
        predictions = []
        for r in res:
            predictions.append(self._predict_single(X, r))
        return np.squeeze(np.array(predictions))

    @property
    def feature_importances_(self):
        assert hasattr(self, "res"), "Must run .fit() first!"
        FI_ = []
        self.SI_ = []
        for estimator in self.best_estimators_:
            fi = estimator.feature_importances_
            FI_.append(fi)
            self.SI_.append(estimator.computeSobol())
        FI_ = np.array(FI_)
        self.SI_ = np.array(self.SI_)

        # weighted average of sobol weighted by error AND explained variance
        # formula for each dim = sum(sobol weight fraction of each component estimator * explain variance ratio for each component)
        try:
            var_weight = self.TT.named_steps.pca.explained_variance_ratio_
        except:
            var_weight = np.ones(self.ntargets) / self.ntargets
        var_weight = var_weight[:, np.newaxis]
        self.sobol_weighted_ = np.sum(self.SI_ * var_weight, axis=0)
        # ***array doesn't have to sum to 1 since the total variance is not captured by the finite set of PCA compoents**. Moreover, it's sum is doubly counted since the sobol indices can sum to more than 1. Thus, the total variance is potentiall greater than the 1-cutoff and less than 1.
        # self.sobol_weighted /= np.sum(self.sobol_weighted)
        return FI_  # return feature importance for each component pce

    def score(self, X, Y):
        Ypred = self.predict(X)
        assert Ypred.shape == Y.shape, "predict and Y shape do not match."
        # MSPREs = [np.mean((1.0 - Ypred[t]/Y[t])**2) for t in range(Y.shape[0])]
        MSPREs = [
            np.mean((Y[t] - Ypred[t]) ** 2) / np.mean(Y[t] ** 2)
            for t in range(Y.shape[0])
        ]
        return -1 * np.mean(MSPREs)

    # add def for fitting multiple for each component for faster fitting

    def compute_cv_score(
        self, X, y, regressor="pce", target_transform=None, scoring="r2"
    ):

        # First clone the surrogate using the best hyper parameters
        n_components = len(self.best_params_)
        reg_custom_list = [regressor for i in range(n_components)]
        reg_param_list = self.best_params_

        if target_transform is None:
            # only works if n_comp is set to exact value
            # will not work if using "auto"
            target_transform = self.TT

        surrogate_clone = MRegressionWrapperCV(
            regressor=reg_custom_list,
            reg_params=reg_param_list,
            custom_params=True,
            target_transform=target_transform,
            target_transform_params={},
            n_jobs=-1,
            verbose=0,
        )

        scores = cross_val_score(surrogate_clone, X, y, scoring=scoring, n_jobs=-1)
        # print("Mean CV score:", scores.mean())

        return scores.mean(), surrogate_clone


class MPCEReg(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        regressor="pce",
        reg_params={"order": 2},
        target_transform=None,
        target_transform_params={},
        scorer="neg_root_mean_squared_error",
        n_jobs=4,
        cv=None,
        mixed=True,
        verbose=1,
    ):
        self.regressor = regressor
        self.reg_params = reg_params
        self.scorer = scorer
        self.n_jobs = n_jobs
        self.cv = cv
        self.target_transform = target_transform
        self.target_transform_params = target_transform_params
        self.mixed = mixed  # to control whether mixed surrogate for each target TBD
        self.verbose = verbose

    def _setupCV(self, shuffle=False, randstate=13):
        if self.cv == None:
            self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)

    def _fit_target_transform(self, Y):
        if self.target_transform is None:
            self.TT = FunctionTransformer(lambda Y: Y)
        else:
            if isinstance(self.target_transform, Pipeline):
                # print("Target Transform is a pipeline object. Cannot set internal parameters just yet.")
                self.TT = self.target_transform
            else:
                self.TT = self.target_transform(**self.target_transform_params)
        self.TT.fit(Y)
        return self

    def fit(self, X, Y):
        self.n, self.dim = X.shape
        self._setupCV()
        self._fit_target_transform(Y)
        Yhat = self.TT.transform(Y)
        if Yhat.ndim == 1:
            Yhat = np.atleast_2d(Yhat).T
        assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
        self.ntargets = Yhat.shape[1]
        self._setupCV()
        if isinstance(self.regressor, str):
            # fits the same regressor to each latent var
            self.res = self.fit_single_reg(X, Yhat)
        elif isinstance(self.regressor, list) and self.mixed == True:
            # fits different regressor to each latent var
            assert (
                len(self.regressor) == self.ntargets
            ), "number of regressors must be same as the targets."
            assert (
                len(self.reg_params) == self.ntargets
            ), "number of regressors must be same as the targets."
            self.res = self.fit_single_reg(X, Yhat)
        # elif isinstance(self.regressor,list) and self.mixed == True:
        # 	self.res = self.fit_single_reg(X,Yhat)
        return self

    def fit_single_reg(self, X, Y, regressor=None, reg_params=None):
        if regressor == None:
            regressor = self.regressor
            reg_params = self.reg_params
        if isinstance(regressor, str):
            # if single string is provided, repeat it for each target
            regressor = [regressor for i in range(self.ntargets)]
            reg_params = [reg_params for i in range(self.ntargets)]
        res = defaultdict(list)
        # with alive_bar(self.ntargets) as bar:
        for i in tqdm(range(self.ntargets)):
            reg = RegressionWrapperCV(
                regressor=[regressor[i]],
                reg_params=[reg_params[i]],
                n_jobs=self.n_jobs,
                scorer=self.scorer,
                cv=self.cv,
                verbose=self.verbose,
            )
            reg.fit(X, Y[:, i])
            res["best_estimators_"].append(reg.best_estimator_)
            res["best_params_"].append(reg.best_params_)
            res["best_scores_"].append(reg.best_score_)
            res["best_overfit_error_"].append(reg.best_overfit_error_)
            res["cv_results_"].append(
                reg.cv_results_
            )  # cv results from best regressor in list
            res["best_index_"].append(reg.best_index_)  # index of best score
            # res['all_cv_results_'].append(reg.all_cv_results_) # all cv results from all regressor in list
            # bar()
        self.__dict__.update(res)
        return res

    def fit_multiple_reg(self, X, Y):
        # wont execute if mixed is True
        if isinstance(self.regressor, list):
            mres = []
            for i, r in enumerate(self.regressor):
                res = self.fit_single_reg(
                    X, Y, regressor=r, reg_params=self.reg_params[i]
                )
                mres.append(res)
        return mres

    def predict(self, X):
        assert hasattr(self, "res"), "Must run fit."
        if isinstance(self.res, dict):
            Ypred = self._predict_single(X, res=self.res)
        elif isinstance(self.res, list):
            Ypred = self._predict_multiple(X, res=self.res)
        return Ypred

    def _predict_single(self, X, res=None):
        # assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
        assert isinstance(
            res, dict
        ), "for single prediction, results must be a dictionary."
        Yhatpred_list = []
        for estimator in self.best_estimators_:
            if X.ndim == 1:
                X = np.atleast_2d(X)
            Yhatpred_list.append(estimator.predict(X))
        Yhatpred = np.array(Yhatpred_list).T
        Ypred = self.TT.inverse_transform(Yhatpred)
        return Ypred

    def _predict_multiple(self, X, res=None):
        # assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
        assert isinstance(
            res, list
        ), "for multiple predictions, results must be a list of dictionaries."
        predictions = []
        for r in res:
            predictions.append(self._predict_single(X, r))
        return np.squeeze(np.array(predictions))

    def feature_importances_(self):
        assert hasattr(self, "res"), "Must run .fit() first!"
        FI_ = []
        for estimator in self.best_estimators_:
            fi = estimator.feature_importances_
            FI_.append(fi)
        return np.array(FI_)

    def score(self, X, Y):
        Ypred = self.predict(X)
        assert Ypred.shape == Y.shape, "predict and Y shape do not match."
        MSPREs = [np.mean((1.0 - Ypred[t] / Y[t]) ** 2) for t in range(Y.shape[0])]
        return np.mean(MSPREs)

    # add def for fitting multiple for each component for faster fitting



import jax
import jax.numpy as jnp
import jax.random as jrandom
import numpy as np
import scipy

from multiindex import MultiIndex

# This needs to run at startup
# https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#Double-(64bit)-precision
jax.config.update('jax_enable_x64', True)

# Polynomials classes
class PolyBase:
    def __init__(self):
        pass
    def Eval1dBasis(self,x):
        # virtual fun to compute basis values at x up to order K
        pass
    def normsq(self,K):
        # compute L2 norm of the basis up to order K
        pass

def Legendre(order,x):
    if order == 0:
        if x.ndim == 0:
            y = 1.0
        else:
            y = jnp.ones(len(x))
    elif order == 1:
        y = x
    elif order == 2:
        y = .5*(3.0*x**2-1)
    elif order == 3:
        y = .5*(5.0*x**3-3*x)
    elif order == 4:
        y = (1./8)*(35.0*x**4 - 30.0*x**2 + 3.0)
    return y
    
class LegPoly(PolyBase):
    '''
    Usage for 1d polynomial basis construction
    L = LegPoly()
    x = np.linspace(-1,1,100)
    L.plotBasis(x,K,normed=True)
    '''
    def __init__(self,domain=[-1.0,1.0]):
        self.domain = domain
        self.a = domain[0]
        self.b = domain[1]
        super().__init__()
    def Eval1dBasis(self,x,K):
        # returns matrix where each column is Li(x)
        self.K = K
        # assert jnp.all(x>=self.a) and jnp.all(x<=self.b), "x is not in the domain."
        # transform x to [-1,1]
        x0 = 2.*(x - 1.*self.a)/(1.*self.b-self.a) - 1.0
        # assert jnp.all(x0>=-1) and jnp.all(x0<=1), "x is not in the domain."
        self.Lis = [Legendre(k,x0) for k in range(self.K+1)]
        return jnp.array(self.Lis) # dim x nx
    def normsq(self,K):
        # compute the squared norm of each basis up to K
        bma = self.domain[1] - self.domain[0]
        normsq = np.array([bma/(2*k+1) for k in range(K+1)])
        self.normsq = normsq
        return normsq

class LegPolyNorm(PolyBase):
    '''
    Usage for 1d polynomial basis construction
    L = LegPoly()
    x = np.linspace(-1,1,100)
    L.plotBasis(x,K,normed=True)
    '''
    def __init__(self,domain=[-1.,1.]):
        self.domain = domain
        self.a = domain[0]
        self.b = domain[1]
        super().__init__()
    def Eval1dBasis(self,x,K):
        # returns matrix where each column is Li(x)
        self.K = K
        # compute norms
        normsq = np.array([(self.b-self.a)/(2*k+1) for k in range(K+1)])
        # check bounds
        # assert jnp.all(x>=self.a) and jnp.all(x<=self.b), "x is not in the domain."
        # transform x to [-1,1]
        x0 = 2.*(x - 1.*self.a)/(1.*self.b-self.a) - 1.0
        # assert jnp.all(x0>=-1) and jnp.all(x0<=1), "x is not in the domain."
        self.Lis = [Legendre(k,x0)/jnp.sqrt(normsq[i]) for i,k in enumerate(range(self.K+1))]
        return jnp.array(self.Lis) # dim x nx
    def normsq(self,K):
        # compute the squared norm of each basis up to K
        bma = self.domain[1] - self.domain[0]
        normsq = np.array([bma/(2*k+1) for k in range(K+1)])
        self.normsq = 1 + 0*normsq
        return normsq

from sklearn.base import BaseEstimator
class PCEBuilder(BaseEstimator):
    def __init__(self,order=1,customM=None,mindex_type='total_order',coef=None,polytype='Legendre',normalized=False):
        # self.dim = dim # no need to initialize with dim
        self.order = order
        self.dim = None
        self.customM = customM
        self.mindex_type = mindex_type
        self.coef = coef
        self.polytype = polytype
        self.normsq = np.array([])
        self.compile_flag = False
        self.mindex = None
        self.normalized = normalized
    def compile(self,dim):
        """Placeholder
        """
        self.dim = dim
        #constructor for different multindices (None, object, and array)
        if self.customM is None:
            self.M = MultiIndex(dim,self.order,self.mindex_type)
        elif isinstance(self.customM,np.ndarray):
            self.dim = self.customM.shape[1] # set dim to match multiindex
            self.M = MultiIndex(self.dim,order=1,mindex_type='total_order') # use default
            self.M.setIndex(self.customM)
            self.order = None # leave blank to indicate custom order
        self.mindex = self.M.index
        self.multiindex = self.mindex
    def fit_transform(self,X):
        """Placeholder
        """
        self.compile(dim=X.shape[1])
        Kmax = jnp.amax(mindex,axis=0)
            # construct and evaluate each basis using mindex
        L = []
        NormSq = []
        self.output_type = None
        if X.ndim == 1: 
            X = jnp.atleast_2d(X)
            self.output_type = 'scalar'
        for i in range(dim):
            # Compute Legendre objects and eval 1d basis
            Leg = LegPoly()
            Li_max = Leg.Eval1dBasis(x=X[:,i],K=Kmax[i])
            L.append(Li_max) # list of size dim of Legendre polys
            NormSq.append(Leg.normsq(Kmax[i])) # norms up to K order

        # start computing products
        Phi = 1.0
        normsq = 1.0
        for di in range(dim):
            Phi = L[di][self.mindex[:,di]] * Phi
            normsq = NormSq[di][self.mindex[:,di]] * normsq
        self.Phi = Phi.T 
        return self.Phi
    def eval(self,c,X=None):
        """Placeholder
        """
        if X is not None:
            self.fit_transform(X)
        
        peval = jnp.dot(self.Phi,c)
        if self.output_type == 'scalar':
            return peval[0]
        else:
            return peval

#########################
######## TEST ###########
#########################

# generate multivariate polynomial
dim = 2
M = MultiIndex(dim=dim,order=1)
mindex = M.index
coef = jnp.array([1.,2,3])
assert len(coef) == len(mindex)
# X = jrandom.uniform(rng,shape=(100,dim))
x = jnp.linspace(-1,1,25)
X = jnp.array([x.flatten() for x in jnp.meshgrid(x,x)]).T

p = PCEBuilder(order=1)
p.fit_transform(X)
def mvp2(c): 
    return p.eval(c=c)
Z = mvp2(coef)

from scipy.optimize import minimize

def fun(c):
    return jnp.mean((Z - mvp2(c=c))**2)

gfun = jax.grad(fun)
gfun_np = lambda c: np.array(jax.jit(gfun)(c))
# gfun_used = gfun_np

fun_and_gfun = jax.value_and_grad(fun)

c0 = np.zeros(3)
optopt = {'gtol': 1e-8, 'disp': False}
res = scipy.optimize.minimize(fun, c0, jac=gfun_np, method='L-BFGS-B', options=optopt)
print(res)

# from mpl_toolkits.mplot3d import Axes3D  
# from matplotlib.pyplot import *
# fig = figure()
# ax = fig.gca(projection='3d')

# surf = ax.scatter(X[:,0],X[:,1],Z)
# ax.set_xlabel('X Label')
# ax.set_ylabel('Y Label')
# ax.set_zlabel('Z Label')
# show()



import numpy as np
import scipy
import time as T
from sklearn.gaussian_process.kernels import (
    ConstantKernel,
    RBF,
    DotProduct,
    WhiteKernel,
)
from sklearn.gaussian_process import GaussianProcessRegressor
import tesuract
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression
from functools import partial
from multiprocessing import Pool
import multiprocessing as mp
from scipy.optimize import fmin_l_bfgs_b
import pickle, pdb
from tqdm import tqdm
from scipy.linalg import sqrtm


from sklearn.linear_model import ElasticNetCV, LassoCV, RidgeCV


def npsolve(K, y):
    """Matrix solver using numpy's linalgebra solve

    This uses numpy's solve mechanism, which automatically checks for positive definiteness

    Parameters
    ----------
    K: np.ndarray
            square matrix to be inverted. Must be positive definite.

    y: np.ndarray
            Must have the same dimensions as the K matrix

    Returns
    -------
    x: np.ndarray
            Returns solution to Kx = y

    """
    # returns inv(K)*y, i.e. solves Kx = y
    return np.linalg.solve(K, y)


def csolve(K, y):
    """Solves the matrix inverse problem using Cholesky decomposition

    Solves Kx = y using the Cholesky decomposition method such that x = inv(K)*y

    Parameters
    ----------
    K: np.ndarray
            square matrix to be inverted. Must be positive definite.

    y: np.ndarray
            Must have the same dimensions as the K matrix

    Returns
    -------
    x: np.ndarray
            Returns solution to Kx = y

    """
    # returns inv(K)*y, i.e. solves Kx = y
    L = scipy.linalg.cholesky(K, lower=True)
    return scipy.linalg.cho_solve((L, True), y), L


def nplml(theta, X, y, kernel, nugget=1e-10):
    """Compute the log marginal likelihood using numpy default solve (slower than cholesky)"""
    kernel_clone = kernel.clone_with_theta(theta)
    K = kernel_clone(X)
    K[np.diag_indices_from(K)] += nugget
    alpha = npsolve(K, y)
    # Compute log-likelihood (compare line 7)
    log_likelihood_dims = -0.5 * np.einsum("ik,ik->k", y, alpha)
    log_likelihood_dims -= 0.5 * np.log(np.linalg.det(K))
    log_likelihood_dims -= K.shape[0] / 2 * np.log(2 * np.pi)
    log_likelihood = log_likelihood_dims.sum(-1)  # sum over dimensions
    return log_likelihood


def clml(theta, X, y, kernel, nugget=1e-10, eval_gradient=False, return_K=False):
    """Compute the log marginal likelihood using Cholesky decomposition

    Modified and adapted from sklearn's GP regression class
    """
    kernel_clone = kernel.clone_with_theta(theta)
    if eval_gradient:
        K, K_gradient = kernel_clone(X, eval_gradient=True)
    else:
        K = kernel_clone(X)
    K[np.diag_indices_from(K)] += nugget
    L = scipy.linalg.cholesky(K, lower=True)
    alpha = scipy.linalg.cho_solve((L, True), y)

    # Compute log-likelihood (compare line 7)
    log_likelihood_dims = -0.5 * np.einsum("ik,ik->k", y, alpha)
    log_likelihood_dims -= np.log(np.diag(L)).sum()  # cholesky includes factor of 1/2
    log_likelihood_dims -= K.shape[0] / 2 * np.log(2 * np.pi)
    log_likelihood = log_likelihood_dims.sum(-1)  # sum over dimensions

    if eval_gradient:  # compare Equation 5.9 from GPML
        tmp = np.einsum("ik,jk->ijk", alpha, alpha)  # k: output-dimension
        tmp -= scipy.linalg.cho_solve((L, True), np.eye(K.shape[0]))[:, :, np.newaxis]
        # Compute "0.5 * trace(tmp.dot(K_gradient))" without
        # constructing the full matrix tmp.dot(K_gradient) since only
        # its diagonal is required
        log_likelihood_gradient_dims = 0.5 * np.einsum("ijl,jik->kl", tmp, K_gradient)
        log_likelihood_gradient = log_likelihood_gradient_dims.sum(-1)
        return log_likelihood, log_likelihood_gradient
    elif return_K:
        return log_likelihood, K
    else:
        return log_likelihood


# minimization solver
def minimize_call(xstart, objfunc, bounds, factr, disp=5):
    """
    Call to minimize in order to use mp within class

    This is a helper function to get multiprocessing to work internally within class. This gets around the pickling error based on the global interpreter lock limitation in Python.
    """
    res = fmin_l_bfgs_b(
        objfunc,
        xstart,
        approx_grad=True,
        bounds=bounds,
        factr=factr,
        pgtol=1e-10,
        disp=disp,
    )
    return res


# utility function for choosing the right starting points and checking bounds
def random_sample_within_bounds(n, bounds, rn=np.random.RandomState(327)):
    """Generate uniform random samples within bounds

    For use in an ensemble Bayesian MCMC method.
    """
    dim = len(bounds)
    samples = []
    low = [b[0] for b in bounds]
    high = [b[1] for b in bounds]
    for d in range(dim):
        usample = rn.rand(n)
        sample_d = usample * (high[d] - low[d]) + low[d]
        samples.append(sample_d)
    samples = np.array(samples).T
    return samples


def check_array_within_bounds(x, bounds):
    """Check that array is within the list of bounds"""
    low, high = np.array(bounds).T
    bounds_check = ((x >= low) & (x <= high)).all()
    return bounds_check


# import emcee
class GaussianProcessPCERegression:
    """Gaussian process regression with polynomial mean function

    This class wraps sklearn's GP regressor class to include a polynomial mean function using the PCEReg from tesuract.

    Parameters
    ----------
    kernel: sklearn.gaussian_process.kernels
        sklearn's GP kernels, e.g., matern, rbf, etc.

    order: int, default=2
        order or degree of the polynomials. By default we start with a quadratic.

    poly_alg: {'monomials','Legendre'}, default='monomials'
        Type of polynomials to use. For higher dimensions, use 'Legendre' since it scales better than the tensor product monomials.

    input_range: list(bounds), default=None
        List of bounds for the input range. If not provided and using Legendre, the input must be scaled from -1 to 1.

    alpha: float
        Unused for now, but I think this was for the pure GP method.

    n_restarts_optimizer: int, default = 0
        How many times to solve the optimization of the log marginal likelihood. If using ElasticNetCV or LassoCV, might want to use the default otherwise it could be very slow. Also, since we are using multiprocessing, it is important to

    random_state: int, default=None
        Random state for choosing the initial guess in the random restarts for optimization

    fit_params: dict, default=None
        Dictionary to give the fit type. E.g., if using sklearn's RidgeCV, you can specify any of the algorithm options. See sklearn's documentation for all the lists of possible fit parameters

    fit_type: {"RidgeCV", "LassoCV", "ElasticNetCV"}, default="RidgeCV"
        fit algorithm used for solving the least squares solution to the polynomial coefficients. Unless the problem is zero noise, it is always better to use a regularized least squares solution. RidgeCV is the fastest of the three, but the others allow l-1 or sparsity constraints which may be useful.

    Examples
    --------
    Given X which is n x d and Y which is n x R and an initial guess of the GP kernel, an a polynomial of order 7 using ElasticNetCV regularization, we can construct the GP+PCE regressor as follows.

    >>> gppce = GPReg(kernel=kernel_guess,order=7,poly_alg='Legendre',input_range=None,n_restarts_optimizer=0,random_state=np.random.RandomState(23423),fit_type='ElasticNetCV',fit_params={'n_jobs':-1})

    We can then fit it by simply calling
    >>> gppce.fit(X,Y)

    and return the predicted polynomial mean and GP standard deviation and covariance

    >>> gppce_mu_pred, gppce_std_pred = gppce.predict(X,return_std=True)
    >>> _, gppce_cov_pred = gppce.predict(X,return_cov=True)

    Notes
    -----
    Due to problems with compatibility with OPENBLAS and the multiprocessing package, it might be necessary to set the environmental variable "export OPENBLAS_NUM_THREADS=1" before running the GPReg fit function.

    """

    def __init__(
        self,
        kernel,
        order=2,
        poly_alg="monomials",
        input_range=None,
        alpha=None,
        n_restarts_optimizer=0,
        random_state=None,
        fit_params=None,
        fit_type="RidgeCV",
    ):
        self.kernel = kernel
        self.order = order
        self.fit_type = fit_type
        self.fit_params = fit_params
        self.poly_alg = poly_alg
        self.input_range = input_range
        self.compile_flag = False
        self.alpha = alpha  # reg param
        self.n_restarts_optimizer = n_restarts_optimizer
        self.random_state = random_state

    def _compile(self, X, y):
        # X must be 2d and y can be 1d
        self.y = y.copy()
        self.X = X.copy()
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
        assert self.X.shape[0] == self.y.shape[0], "shape mistmatch."
        if self.order == 0:
            self.Xhat = self.X
        elif self.order > 0:
            if self.poly_alg == "Legendre":
                self.pce = tesuract.PCEBuilder(
                    order=self.order, input_range=self.input_range
                )  # store_phi=True to keep in the pce object
            if self.poly_alg == "monomials":
                self.pce = PolynomialFeatures(degree=self.order)
            self.Xhat = self.pce.fit_transform(self.X)
            self.coef_size = self.Xhat.shape[1]
        self.theta_size = len(self.kernel.theta)
        self.feature_labels = [
            "theta_" + str(t + 1) for t in range(self.theta_size)
        ] + ["coef_" + str(c) for c in range(self.coef_size)]
        self.theta_bounds = [tuple(bi) for bi in self.kernel.bounds]
        self.compile_flag = True
        # self.coef_bounds = [(-np.inf,np.inf) for c in range(self.coef_size)]
        self.coef_bounds = [(-1e1, 1e1) for c in range(self.coef_size)]
        self.all_bounds = self.theta_bounds + self.coef_bounds
        return self

    def log_marginal_likelihood(self, theta, nugget=1e-8, return_coef=False):
        """Compute the log marginal likelihood of the data using the kernel hyper parameters and polynomial

        We use the formulation as outlined in https://www.mathworks.com/help/stats/exact-gpr-method.html. Note there is some uncertainty and confusion about the sigma variable in matlab's documentation and the white noise kernel parameter in sklearn's formulation. In this code and in sklearn's, if we add a white noise kernel to the full kernel, the log marginal likelihood is formulated the same as MATLAB's. The difference is in the prediction. Here, since we are separating out the stochastic/ GP from the polynomial model, it may not matter, but if we were to use MATLAB's posterior prediction, we have to modify the predict function.

        The log marginal likelihood can be written purely as a function of the kernel hyper parameters since the polynomial coefficient sub-problem can be written as a least squares problem with an exact OLS solution. This means that we do not have to search over the space of polynomial coefficients, avoiding, in part, the curse of dimensionality.

        Parameters
        ----------
        theta : np.ndarray
            kernel hyper-paramaters as computed using sklearn's kernel structure. Note that this structure is very confusing, so we try to leave it untouched to avoid bugs.

        nugget : float, default=1e-8
            Component to add to diagonal to avoid convergence issues with condition numbers. Make this number larger if having problems with Cholesky solves.

        return_coef : bool, default=False
            returns the polynomial coefficients. For use in debugging, but internally it is needed for the alternating minimization approach outlined in the MATLAB link above where we then solve the polynomial least squares problem.

        Returns
        -------
        float
            returns the scalar number which represents the log marginal likelihood of the data, given the polynomial and kernel theta hyper-parameters

        """
        # compute pce coefficient first and then use the coef to evaulate lml
        kernel_clone = self.kernel.clone_with_theta(theta)
        K = kernel_clone(self.X)
        K[np.diag_indices_from(K)] += nugget
        L = scipy.linalg.cholesky(K, lower=True)
        H = self.Xhat
        # KiH = np.linalg.solve(K,H) #scipy.linalg.cho_solve((L, True), H)
        # HTKiH_palpha = np.dot(H.T,KiH) + self.alpha*np.eye(self.coef_size)
        # Kiy = np.linalg.solve(K,self.y) #scipy.linalg.cho_solve((L,True),self.y)
        # HTKiy = np.dot(H.T,Kiy)
        # beta = np.linalg.solve(HTKiH_palpha,HTKiy)
        # beta_bar = beta.mean(axis=1)
        # coef = beta_bar
        # print("coef1",coef[:5])

        ############ new method
        Xtilde = np.linalg.solve(L, H)
        if self.y.ndim == 1:
            ytilde = np.linalg.solve(L, self.y).flatten()
        else:
            ytilde = np.linalg.solve(L, self.y.mean(axis=1)).flatten()

        # Default linear solver is ridge regression cv in sklearn with 25 alphas
        ridgecv_fit_params = {"alphas": np.logspace(-3, 3, 50)}
        self.fit_params = {}  # default empty fit params
        if self.fit_type is None:
            self.fit_type = "RidgeCV"
            self.fit_params = ridgecv_fit_params

        if self.fit_type == "RidgeCV":
            regr = RidgeCV(fit_intercept=False, **self.fit_params)
        elif self.fit_type == "ElasticNetCV":
            regr = ElasticNetCV(fit_intercept=False, **self.fit_params)
        elif self.fit_type == "LassoCV":
            regr = LassoCV(fit_intercept=False, **self.fit_params)

        regr.fit(Xtilde, ytilde)
        coef = regr.coef_
        #######################

        self.y_minus_mu = self.y - np.dot(self.Xhat, coef)[:, np.newaxis]
        assert (
            self.compile_flag == True
        ), "Must compile first to precompute polynomial transform and save X,y data."
        # assert len(theta) == self.theta_size, "theta wrong length."
        # Compute log-likelihood (compare line 7)
        Kiymu = scipy.linalg.cho_solve((L, True), self.y_minus_mu)
        log_likelihood_dims = -0.5 * np.einsum("ik,ik->k", self.y_minus_mu, Kiymu)
        log_likelihood_dims -= np.log(np.diag(L)).sum()  # cholesky w factor of 1/2
        log_likelihood_dims -= K.shape[0] / 2 * np.log(2 * np.pi)
        log_likelihood = log_likelihood_dims.sum(-1)  # sum over dimensions
        # loglike = clml(theta,self.X,self.y_minus_mu,self.kernel,nugget=nugget,eval_gradient=False)
        if return_coef:
            return log_likelihood, coef
        else:
            return log_likelihood

    def _objfunc(self, theta, minimize=True):
        """Wrapper for the log marginal likelihood to turn the problem into a minimization instead of a maximization.

        Parameters
        ----------
        theta : np.ndarray
            kernel hyper-parameters. Note that the polynomial coefficients are completely determined by the kernel hyper-parameters using an OLS analytic solution (see the MATLAB link above for the formulation).

        minimize : bool, default=True
            Turns the log marginal likelihood objective into a minimization problem by adding a factor of -1 if True, and 1 if False (for maximization). The latter may be used for a Bayesian formulation where the log likelihood is needed in its maximization form.

        Returns
        -------
        float
            Returns the log marginal likelihood
        """
        mfact = 1.0
        if minimize:
            mfact = -1.0
        lml = self.log_marginal_likelihood(theta)
        lml = mfact * lml
        return lml

    def fit(self, X, y, **fit_params):
        """Main fit routine for determining the GP kernel hyper-parameters and polynomial coefficients for the stochastic/ repeated data X, Y

        Fitting a Polynomial + GP to the data pair X,y where y can have repeated elements. In general, it is difficult to deal with repeated data in a standard sense since this means your data matrix would have to have repeated rows, thus reducing the conditioning of your system and making the GP approach much more difficult (it scales like O(n^3) where n is the total number of samples). This routine essentially fits one GP model using a log marginal likelihood that averages the error out over the repeated samples. If you were to just fit a deterministic regression model to the averaged data, you would obtain a good mean approximation but loose information about the stochastic component. This helps avoid that.

        Parameters
        ----------
        X : np.ndarray
            Feature space data matrix, n x feature_dimensions, where n is the number of spatial samples
        y : np.ndarray
            Can be a matrix of size n x R, where R is the number of repeated outputs. Since this is meant for stochastic code, each realization of the simulation can result in slightly different target values so y can be a matrix, where the columns are the repeated samples.

        Returns
        -------
        self
            returns the GPReg object itself

        """
        self._compile(X, y)
        opt_args = {"factr": 1e7, "disp": False}
        myfun = partial(
            minimize_call, objfunc=self._objfunc, bounds=self.theta_bounds, **opt_args
        )
        nsolves = self.n_restarts_optimizer + 1
        # xstarts = [np.zeros(len(self.all_bounds)) for i in range(nsolves)]
        if self.random_state is None:
            self.random_state = np.random.RandomState(997)
        ##########
        start_bounds = self.theta_bounds
        xstarts = random_sample_within_bounds(
            nsolves, start_bounds, rn=self.random_state
        )
        # add solver for pce coef only to improve problem of poor fit when data not scaled.
        ##########
        # xstarts = [2*self.random_state.rand(len(self.all_bounds))-1 for i in range(nsolves)]
        # print(xstarts)
        print("Starting parallel optimization ...")
        start = T.time()
        with mp.get_context("fork").Pool() as pool:
            res = pool.map(myfun, [x for x in xstarts])
        end = T.time()
        # print("Total time for optimization is %.5f seconds." %(end-start))

        xvalues = [r[0] for r in res]
        fvalues = [r[1] for r in res]
        min_index = np.argmin(fvalues)
        self.max_logp_ = -1 * fvalues[min_index]
        self.xmap_ = xvalues[min_index]
        self.theta_ = self.xmap_  # separate theta
        logl, self.coef_ = self.log_marginal_likelihood(
            self.theta_, return_coef=True
        )  # self.xmap_[self.theta_size:]   # separate pce coef
        return self

    def predict(
        self,
        X,
        return_pce_only=False,
        return_std=False,
        return_cov=False,
        return_samples=None,
    ):
        # plot the test prediction using the GP PCE model
        y_mu_gppce_pred = self.pce.polyeval(X, c=self.coef_)
        kernel_opt_ = self.kernel.clone_with_theta(self.theta_)
        rn = self.random_state
        gp = GaussianProcessRegressor(kernel=kernel_opt_, random_state=rn)
        gp_mu_resid, gp_std_resid = gp.predict(X, return_std=True)
        _, gp_cov_resid = gp.predict(X, return_cov=True)
        if return_std is True:
            return y_mu_gppce_pred + gp_mu_resid, gp_std_resid
        elif return_cov is True:
            return y_mu_gppce_pred + gp_mu_resid, gp_cov_resid
        elif return_pce_only is True:
            return y_mu_gppce_pred, self.pce
        if isinstance(return_samples, int):
            nsamples = return_samples
            gp_resid_samples = gp.sample_y(X, nsamples)
            return y_mu_gppce_pred[:, np.newaxis] + gp_resid_samples
        else:
            return y_mu_gppce_pred + gp_mu_resid


# metrics for comparing outputs
def kl_mvn(m0, S0, m1, S1):
    """
    Kullback-Liebler divergence from Gaussian pm,pv to Gaussian qm,qv.
    Also computes KL divergence from a single Gaussian pm,pv to a set
    of Gaussians qm,qv.
    Diagonal covariances are assumed.  Divergence is expressed in nats.
            ONLY FOR DIAGONAL COV
    - accepts stacks of means, but only one S0 and S1

    From wikipedia
    KL( (m0, S0) || (m1, S1))
             = .5 * ( tr(S1^{-1} S0) + log |S1|/|S0| +
                              (m1 - m0)^T S1^{-1} (m1 - m0) - N )
    """
    # store inv diag covariance of S1 and diff between means
    N = m0.shape[0]
    iS1 = np.linalg.inv(S1)
    diff = m1 - m0

    # kl is made of three terms
    tr_term = np.trace(iS1 @ S0)
    # det_term  = np.log(np.linalg.det(S1)/np.linalg.det(S0))
    # det_term = np.sum(np.log(S1)) - np.sum(np.log(S0))
    det_term = np.sum(np.log(np.diag(S1) / np.diag(S0)))
    quad_term = diff.T @ np.linalg.inv(S1) @ diff  # np.sum( (diff*diff) * iS1, axis=1)
    # print(tr_term,det_term,quad_term)
    return 0.5 * (tr_term + det_term + quad_term - N)


def Wasserstein2(mu1, cov1, mu2, cov2):
    """Wasserstein or earth mover distance

    For large matrices that have very small eigenvalues, this may not work well.
    """
    mu_sse = np.sum((mu1 - mu2) ** 2)
    trC1 = np.trace(cov1)
    trC2 = np.trace(cov2)
    sqrtS1 = sqrtm(cov1)
    B2 = trC1 + trC2 - 2 * np.trace(sqrtm(sqrtS1 @ cov2 @ sqrtS1))
    return mu_sse + B2



import numpy as np 
import pypce
from collections import defaultdict
import time as T
import matplotlib.pyplot as mpl

import warnings
 
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.estimator_checks import check_estimator 
from sklearn.utils.validation import check_X_y, check_array
from sklearn.model_selection import GridSearchCV, ParameterGrid, KFold, cross_val_score, cross_validate
from sklearn.ensemble import RandomForestRegressor
from sklearn.preprocessing import FunctionTransformer
from sklearn.metrics import make_scorer
from sklearn.decomposition import PCA
   
class RegressionWrapperCV(BaseEstimator):
	def __init__(self,regressor='pce',reg_params={},scorer='neg_root_mean_squared_error',n_jobs=1,verbose=1,cv=None):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.verbose = verbose
		self.cv = cv
		# self.__dict__.update(reg_params)
	def _setup_cv(self):
		if self.cv == None:
			self.cv = KFold(n_splits=5)
	def _reformat_grid(self,params):
		try: 
			# if parameter grid is in the form of grid search cv
			ParameterGrid(params)
			params_cv = params
		except:
			# convert reg_params to form that can be used by grid search
			# for a single grid point
			params_cv = defaultdict(list)
			for key,item in params.items():
				params_cv[key].append(item)
		return params_cv
	def _model_factory(self,regressor):
		if regressor == 'pce':
			return pypce.pcereg()
		if regressor == 'randforests':
			return RandomForestRegressor()
	def fit(self,X,y):
		self._setup_cv()
		if isinstance(self.regressor,str):
			model = self._model_factory(regressor=self.regressor)
			reg_params_cv = self._reformat_grid(self.reg_params)
			GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
			GridSCV.fit(X,y)
			self.best_estimator_ = GridSCV.best_estimator_
			self.best_params_ = GridSCV.best_params_
			self.best_score_ = GridSCV.best_score_
			self.best_overfit_error_ = self.overfit_score(GridSCV)
			self.cv_results_ = GridSCV.cv_results_
		if isinstance(self.regressor,list):
			self.fit_multiple_reg(X,y)
		return self
	def fit_multiple_reg(self,X,y):
		if isinstance(self.regressor,list):
			n_regressors = len(self.regressor)
			self.best_estimators_ = [None]*n_regressors
			self.all_best_params_ = [None]*n_regressors
			self.best_scores_ = np.zeros(n_regressors)
			self.best_overfit_errors_ = [None]*n_regressors
			self.all_cv_results_ = [None]*n_regressors
			assert isinstance(self.reg_params,list), "parameters must also be a list"
			assert len(self.reg_params) == n_regressors, "length of parameters and regressors must match."
			for i,R in enumerate(self.regressor):
				print("Fitting regressor ", R)
				model = self._model_factory(regressor=R)
				reg_params_cv = self._reformat_grid(self.reg_params[i])
				GridSCV = GridSearchCV(model, reg_params_cv, scoring=self.scorer, n_jobs=self.n_jobs, cv=self.cv, verbose=self.verbose, return_train_score=True)
				GridSCV.fit(X,y)
				self.best_estimators_[i] = GridSCV.best_estimator_
				self.all_best_params_[i] = GridSCV.best_params_
				self.best_scores_[i] = GridSCV.best_score_
				self.best_overfit_errors_[i] = self.overfit_score(GridSCV)
				self.all_cv_results_[i] = GridSCV.cv_results_
			self.best_index_ = np.argmax(self.best_scores_)
			self.best_estimator_ = self.best_estimators_[self.best_index_]
			self.best_params_ = self.all_best_params_[self.best_index_]
			self.best_score_ = self.best_scores_[self.best_index_]
			self.best_overfit_error_ = self.best_overfit_errors_[self.best_index_]
			self.cv_results_ = self.all_cv_results_[self.best_index_]
		return self
	def predict(self,X):
		return self.best_estimator_.predict(X)
	def overfit_score(self,GridSCV):
		''' Takes a grid search cv object and returns overfit score '''
		# assert hasattr(self,'GridSCV'), "Must run fit function."
		best_index_ = GridSCV.best_index_
		mean_train_score = GridSCV.cv_results_['mean_train_score']
		mean_test_score = GridSCV.cv_results_['mean_test_score']
		return mean_train_score[best_index_] - mean_test_score[best_index_]

from tqdm import tqdm
class MRegressionWrapperCV(BaseEstimator, RegressorMixin):
	def __init__(self, 
				regressor='pce', reg_params={'order':2},
				target_transform=None,
				target_transform_params={},
				scorer='neg_root_mean_squared_error',
				n_jobs=4,cv=None,
				mixed=True,
				verbose=1):
		self.regressor = regressor
		self.reg_params = reg_params
		self.scorer = scorer
		self.n_jobs = n_jobs
		self.cv = cv
		self.target_transform = target_transform
		self.target_transform_params = target_transform_params
		self.mixed = mixed # to control whether mixed surrogate for each target TBD
		self.verbose = verbose
	def _setupCV(self, shuffle=False, randstate=13):
		if self.cv == None:
			self.cv = KFold(n_splits=5)  # ,shuffle=True,random_state=13)
	def _fit_target_transform(self,Y):
		if self.target_transform is None:
			self.TT = FunctionTransformer(lambda Y: Y)
		else:
			self.TT = self.target_transform(**self.target_transform_params)
		self.TT.fit(Y)
		return self
	def fit(self,X,Y):
		self.n, self.dim = X.shape
		self._setupCV()
		self._fit_target_transform(Y)
		Yhat = self.TT.transform(Y)
		if Yhat.ndim == 1: Yhat = np.atleast_2d(Yhat).T
		assert len(Yhat) == self.n, "mistmatch in no of samples btwn X and Y."
		self.ntargets = Yhat.shape[1]
		self._setupCV()
		if isinstance(self.regressor,str):
			self.res = self.fit_single_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == False:
			self.res = self.fit_multiple_reg(X,Yhat)
		elif isinstance(self.regressor,list) and self.mixed == True:
			self.res = self.fit_single_reg(X,Yhat)
		return self
	def fit_single_reg(self, X, Y, regressor=None,reg_params=None):
		if regressor == None:
			regressor = self.regressor
			reg_params = self.reg_params
		res = defaultdict(list)
		for i in tqdm(range(self.ntargets)):
			reg = RegressionWrapperCV(
				regressor=regressor,reg_params=reg_params,
				n_jobs=self.n_jobs, scorer=self.scorer, cv=self.cv, verbose=self.verbose)
			reg.fit(X, Y[:, i])
			res['best_estimators_'].append(reg.best_estimator_)
			res['best_params_'].append(reg.best_params_)
			res['best_scores_'].append(reg.best_score_)
			res['best_overfit_error_'].append(reg.best_overfit_error_)
		self.__dict__.update(res)
		return res
	def fit_multiple_reg(self, X, Y):
		# wont execute if mixed is True
		if isinstance(self.regressor,list):
			mres = []
			for i,r in enumerate(self.regressor):
				res = self.fit_single_reg(X,Y,
								regressor=r,
								reg_params=self.reg_params[i])
				mres.append(res)
		return mres
	def predict(self,X):
		assert hasattr(self,'res'), "Must run fit."
		if isinstance(self.res,dict):
			Ypred = self._predict_single(X,res=self.res)
		elif isinstance(self.res,list):
			Ypred = self._predict_multiple(X,res=self.res)
		return Ypred
	def _predict_single(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,dict), "for single prediction, results must be a dictionary."
		Yhatpred_list = []
		for estimator in self.best_estimators_:
			Yhatpred_list.append(estimator.predict(X))
		Yhatpred = np.array(Yhatpred_list).T
		Ypred = self.TT.inverse_transform(Yhatpred)
		return Ypred
	def _predict_multiple(self,X,res=None):
		# assert isinstance(self.res,dict), "Must pass string as regressor OR list with mixed=True, otherwise, predict is ambiguous"
		assert isinstance(res,list), "for multiple predictions, results must be a list of dictionaries."
		predictions = []
		for r in res:
			predictions.append(self._predict_single(X,r))
		return np.squeeze(np.array(predictions))
	def feature_importances_(self):
		assert hasattr(self,'res'), "Must run .fit() first!"
		FI_ = []
		for estimator in self.best_estimators_:
			fi = estimator.feature_importances_
			FI_.append(fi)
		return np.array(FI_)

class PCATargetTransform(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=2, cutoff=1e-2, svd_solver='arpack'):
        self.n_components = n_components
        self.cutoff = cutoff
        self.svd_solver = svd_solver
        self.K = None
    def fit(self, Y):
        self.n, self.d = Y.shape
        if self.n_components == 'auto':
            self._compute_K(Y)
        if isinstance(self.n_components, int):
            self.K = self.n_components
            if self.n_components > self.d:
                warnings.warn("No of components greater than dimension. Setting to maximum allowable.")
                self.n_components = self.d
        assert self.K is not None, "K components is not defined or being set properly."
        self.pca = PCA(n_components=self.K, svd_solver=self.svd_solver)
        self.pca.fit(Y)
        self.cumulative_error = np.cumsum(self.pca.explained_variance_ratio_)
        return self
    def fit_transform(self, Y):
        self.fit(Y)
        return self.pca.transform(Y)
    def transform(self, Y):
        assert hasattr(self, 'pca'), "Perform fit first."
        return self.pca.transform(Y)
    def inverse_transform(self, Yhat):
        return self.pca.inverse_transform(Yhat)
    def _compute_K(self, Y, max_n_components=50):
        ''' automatically compute number of PCA terms '''
        pca = PCA(n_components=min(max_n_components, self.d), svd_solver=self.svd_solver)
        pca.fit(Y)
        cumulative_error = np.cumsum(pca.explained_variance_ratio_)
        print(cumulative_error)
        loc = np.where(1 - cumulative_error <= self.cutoff)[0]
        if loc.size == 0:
        	warnings.warn("Cutoff may be too big. Setting K = 16")
        	self.K = 16
        elif loc.size > 0:
	        self.K = loc[0]


# data
n, d = 30, 2
rn = np.random.RandomState(45)
X = 2 * rn.rand(n, d) - 1
y = 3 + 2 * .5 * (3 * X[:, 0] ** 2 - 1) + X[:, 1] * .5 * (3 * X[:, 0] ** 2 - 1)
y2 = 1 + 2 * .5 * (3 * X[:, 1] ** 2 - 1) + 3 * X[:, 0] * .5 * (3 * X[:, 1] ** 2 - 1)
Y = np.vstack([y, y2]).T  # test for multioutput regressor
c_true = np.array([3., 0., 0., 2., 0., 0., 0., 1., 0., 0., ])


# single parameter test
reg_params = {'order': 3, 'mindex_type': 'total_order', 'fit_type': 'linear'}
RW = RegressionWrapperCV(regressor='pce',reg_params=reg_params)
# RW.fit(X,y)

# pce grid search cv test
pce_param_grid = [{'order': [1, 2, 3, 4],
	 'mindex_type': ['total_order', 'hyperbolic', ],
	 'fit_type': ['LassoCV', 'linear', 'ElasticNetCV']}]
RW2 = RegressionWrapperCV(regressor='pce',reg_params=pce_param_grid,n_jobs=6)


# random forest fit
rf_param_grid = {'n_estimators': [500,1000],
				 'max_features': ['auto'],
				 'max_depth': [5,10,15]
				 }
RW3 = RegressionWrapperCV(regressor='randforests', reg_params=rf_param_grid,n_jobs=8)

# random forest fit
rf_param_grid2 = {'n_estimators': [1000],
				 'max_features': ['log2'],
				 'max_depth': [5]
				 }
RW4 = RegressionWrapperCV(regressor='randforests',reg_params=rf_param_grid2,n_jobs=8)

# compute cv score for single target and list of regressors
multreg = RegressionWrapperCV(
		regressor=['pce','pce','randforests','randforests'],
		reg_params=[pce_param_grid,reg_params,rf_param_grid,rf_param_grid2],
		n_jobs=6)

# fit multiple target with single type
mtarg1a = MRegressionWrapperCV(
			regressor='pce',
			reg_params=pce_param_grid,
			n_jobs=8)

# fit multiple target with single type
mtarg1b = MRegressionWrapperCV(
			regressor='randforests',
			reg_params=rf_param_grid,
			n_jobs=8)

# fit multiple target with type as single element list
mtarg2 = MRegressionWrapperCV(
			regressor=['pce'],
			reg_params=[pce_param_grid],
			n_jobs=8)


# fit multiple target with more than one type and mixed basis
mtarg3 = MRegressionWrapperCV(
			regressor=['pce','randforests'],
			reg_params=[pce_param_grid,rf_param_grid],
			n_jobs=8,
			mixed=True)

# fit multiple target with more than one type
# returns a list of models, not suitable for cross-validation
# probably won't include because we can't use validation score from sklearn
mtarg4 = MRegressionWrapperCV(
			regressor=['pce','randforests'],
			reg_params=[pce_param_grid,rf_param_grid],
			n_jobs=4,
			mixed=False)

# cross validation test
def multi_target_metric(y_true,y_pred):
	if y_true.ndim == 2 and y_true.ndim == 2: 
	    error_array =  np.linalg.norm(y_true - y_pred,axis=1)/np.linalg.norm(y_true,axis=1)
	    return error_array.mean()
	else:
		return np.linalg.norm(y_true - y_pred)/np.linalg.norm(y_true)
scorer = make_scorer(multi_target_metric,greater_is_better=False)
cv_outer = KFold(n_splits=5)
# cv_results = cross_validate(mtarg1a,X,Y,scoring=scorer,cv=cv_outer,verbose=1,return_train_score=True)

# grid search cv
param_grid = [
			  {'regressor': ['pce'], 'reg_params': [pce_param_grid]},
			  {'regressor': ['randforests'], 'reg_params': [rf_param_grid]}
			 ]
gsCV = GridSearchCV(MRegressionWrapperCV(n_jobs=8,verbose=0), param_grid, 
					scoring=scorer, cv=cv_outer, 
					verbose=1, return_train_score=True)
start = T.time()
# gsCV.fit(X,Y)
end = T.time()
print("Total time is %.5f seconds" %(end-start))

################################################################
# PCA Target Transform
################################################################
print("Performing PCA transform...")
# test target transform
# Load the data
data_fields = ['heat_flux_wall','df_pressure','df_wall-heat-flux',
			   'df_wall-C_f','df_wall-C_h','df_wall-friction-velocity',
			   'df_wall-tau-mag']
# for d in data_fields[1:]:
# 	test = np.load(datadir + 'data/' + d + '.npy')
# 	np.save('Y_' + d + '.npy', test)
data_field = data_fields[5]
# for data_field in data_fields:
print(data_field)
datadir = '/Users/kchowdh/Research/sparc_tests/'
X = np.load(datadir + 'data/X_samples.npy')
x_domain = np.load(datadir + 'data/x_domain.npy')
Y = np.load(datadir + 'data/Y_' + data_field + '.npy')

Q = [.05,.5,.95]

# Scale target Y (use inverse transform for new data)
# Each row is a functional output, so we should scale each row
target_scaler = pypce.preprocessing.MinMaxTargetScaler(target_range=(0, 1))
Y_scaled = target_scaler.fit_transform(Y)
Y_q = np.quantile(Y_scaled,Q,axis=0)

# Lower and Upper Bound for features in X (for future scaling)
X_LB = np.array([.85, .85, .85, 0.7, 0.8, 0.3, 0.7, 0.0784, 0.38, 0.31, 1.19, 1.05])
X_UB = np.array([1.15, 1.15, 1.15, 1.0, 1.2, 0.7, 1.0, 0.1024, 0.42, 0.40, 1.31, 1.45])
X_bounds = list(zip(X_LB,X_UB))

X_col_names = ['dens_sc', 'vel_sc', 'temp_sc',
               'sigk1', 'sigk2', 'sigw1', 'sigw2',
               'beta_s', 'kap', 'a1', 'beta1r', 'beta2r']

# Scale features X (use inverse transform for new data)
feature_scaler = pypce.preprocessing.DomainScaler(dim=12, input_range=X_bounds, output_range=(-1, 1))
X_scaled = feature_scaler.fit_transform(X)

# pca_params={'n_components':'auto','cutoff':1e-3}
pca_params={'n_components':8}
pca = PCATargetTransform(**pca_params)
pca.fit(Y_scaled)
Yhat_pca = pca.transform(Y_scaled)
print('n components = ', Yhat_pca.shape[1])
Y_pca_recon = pca.inverse_transform(Yhat_pca)
Y_pca_q = np.quantile(Y_pca_recon,Q,axis=0)
pca_recon_error = np.mean((Y_scaled - Y_pca_recon)**2,axis=1)

# UMAP test
import umap
print("Running UMAP...")
reducer = umap.UMAP(n_components=2)
embedding = reducer.fit(Y_scaled)
Yhat_umap = embedding.transform(Y_scaled)
Y_umap_recon = embedding.inverse_transform(Yhat_umap)
umap_recon_error = np.mean((Y_scaled - Y_umap_recon)**2,axis=1)

# # fit multiple target with type as single element list
# HFreg = MRegressionWrapperCV(
# 			regressor=['pce'],
# 			reg_params=[pce_param_grid],
# 			n_jobs=8)
# # HFreg.fit(X_scaled,Yhat_pca)
# # Yhat_pred = HFreg.predict(X_scaled)
# # Y_pce_recon = pca.inverse_transform(Yhat_pred)
# # Y_pce_q = np.quantile(Y_pce_recon,Q,axis=0)


# # fit multiple target with type as single element list
# HFreg2 = MRegressionWrapperCV(
# 			regressor=['randforests'],
# 			reg_params=[rf_param_grid],
# 			target_transform=PCATargetTransform,
# 			target_transform_params=pca_params,
# 			n_jobs=8)
# HFreg2.fit(X_scaled,Y_scaled)
# Y_pce_recon2 = HFreg2.predict(X_scaled)
# Y_pce_q = np.quantile(Y_pce_recon2,Q,axis=0)
# print(HFreg2.best_scores_)

# # PLOTTING

# def plot_feature_importance(S,feature_labels,extra_line_plot=None):
#     assert isinstance(S,np.ndarray), "S must be a numpy array"
#     if S.ndim == 1:
#         ntargets = 1
#         ndim = len(S)
#         S = np.atleast_2d(S)

#     ntargets, ndim = S.shape
#     # normalize across columns (if not already)
#     S = S / np.atleast_2d(S.sum(axis=1)).T

#     # plot sobol indices as stacked bar charts

#     import matplotlib.pyplot as plt
#     import matplotlib._color_data as mcd
#     import seaborn as sns

#     xkcd_colors = []
#     xkcd = {name for name in mcd.CSS4_COLORS if "xkcd:" + name in mcd.XKCD_COLORS}
#     for j, n in enumerate(xkcd):
#         xkcd = mcd.XKCD_COLORS["xkcd:" + n].upper()
#         xkcd_colors.append(xkcd)

#     Ps = []
#     sns.set_palette(sns.color_palette("Paired", 12))
#     plt.figure(figsize=(20, 9))
#     bottom = np.zeros(ntargets)
#     for ii in range(ndim):
#         ptemp = plt.bar(np.arange(1, 1 + ntargets), S[:, ii], bottom=bottom, width=.25)
#         bottom = bottom + S[:, ii]  # reset bottom to new height
#         Ps.append(ptemp)
#     plt.ylabel('Sobol Scores')
#     plt.ylim([0, 1.1])
#     # plt.title('Sobol indices by pca mode')
#     # plt.xticks(t, ('PCA1','PCA2','PCA3','PCA4','PCA5','PCA6'))
#     # X_col_names = ['dens_sc', 'vel_sc', 'temp_sc',
#     #            'sigk1', 'sigk2', 'sigw1', 'sigw2',
#     #            'beta_s', 'kap', 'a1', 'beta1r', 'beta2r']
#     plt.legend(((p[0] for p in Ps)), (l for l in feature_labels),
#            fancybox=True, shadow=True,
#            loc='upper center', bbox_to_anchor=(0.5, 1.05), ncol=ndim)

#     # plot explained variance
#     if extra_line_plot is not None:
#         assert len(extra_line_plot) >= ntargets, "extra info must be the same length as the targets."
#         plt.plot(np.arange(1, 1 + ntargets), extra_line_plot[:ntargets], '--ok')
#     return plt

# plt = plot_feature_importance(HFreg2.feature_importances_(),X_col_names,pca.cumulative_error)
# plt.savefig(data_field + '_features.png')

# fig, ax = mpl.subplots(1,1,figsize=(20,9))
# x = 1.0*np.arange(len(Y_pca_q[0]))
# ax.fill_between(x=x/np.amax(x),y1=Y_pca_q[0],y2=Y_pca_q[2],color='b',alpha=.5)
# ax.plot(x/np.amax(x),Y_pca_q[1],'-w',alpha=.25)
# ax.fill_between(x=x/np.amax(x),y1=Y_pce_q[0],y2=Y_pce_q[2],color='r',alpha=.15)
# ax.plot(x/np.amax(x),Y_pce_q[1],'--r',alpha=.25)
# ax.set_title(data_field)

# fig.savefig(data_field + '_quantiles.png')
# # plt.show()
# # fig.show()








import numpy as np
import pdb, warnings, pickle
from numpy.polynomial.legendre import leggauss
from numpy.polynomial.hermite_e import hermegauss
import itertools
from scipy.special import comb

"""
This code has been heavily modified and adapted from UQTk (https://github.com/sandialabs/UQTk) and Chaospy (This code has been adapted from Chaospy (https://github.com/jonathf/chaospy) using custom multiindex generation and circumventing the use of external libraries like numpoly)

The modifications are substantial to (1) allow for quadrature generation using pure Python code and (2) remove the need for any third party extraneous polynomial libraries other than numpy.  
"""


class QuadBase:
    def __init__(self, nquad):
        self.nquad = nquad

    def get1dQuad(self):
        pass


class LegendreQuad(QuadBase):
    def __init__(self, nquad=2):
        super().__init__(nquad)

    def get1dQuad(self, nquad=None):
        if nquad is not None:
            self.nquad = nquad
        x, w = leggauss(self.nquad)
        # rescale weights to sum to 1
        w = w / 2.0
        return x, w


class HermiteQuad(QuadBase):
    """normalized"""

    def __init__(self, nquad=2):
        super().__init__(nquad)

    def get1dQuad(self, nquad=None):
        if nquad is not None:
            self.nquad = nquad
        x, w = hermegauss(self.nquad)
        return x, w  # add a factor of (2*np.pi)**-.5 to normalize each dimension


class ClenshawCurtis(QuadBase):
    def __init__(self, nquad=2):
        super().__init__(nquad)

    def _get1dQuad(self, nquad=None):
        """old"""
        if nquad is not None:
            self.nquad = nquad
        if self.nquad == 1:
            return np.array([0.0]), np.array([2.0])
        else:
            n = self.nquad
            x = np.cos(np.pi * (n - 1 - np.arange(n)) / (n - 1))
            w = np.ones(len(x))
            for i in range(n):
                theta = i * np.pi / (n - 1)
                for j in range(1, int(0.5 * (n - 1) + 1)):
                    if 2 * j == n - 1:
                        f = 1.0
                    else:
                        f = 2.0
                    w[i] -= f * np.cos(2.0 * j * theta) / (4.0 * j ** 2 - 1)
            w[0] /= n - 1
            w[1:-1] = 2 * w[1:-1] / (n - 1)
            w[-1] *= 1.0 / (n - 1)
            return x, w

    def get1dQuad(self, nquad=None):
        """from chaospy"""
        if nquad is not None:
            self.nquad = nquad
        degree = self.nquad
        n = self.nquad
        if n == 1:
            points = np.array([0.0])
            weights = np.array([2.0])
        else:
            points = -np.cos((np.pi * np.arange(n)) / (n - 1))
            if n == 2:
                weights = np.array([1.0, 1.0])
            else:
                n -= 1
                N = np.arange(1, n, 2)
                length = len(N)
                m = n - length
                v0 = np.concatenate(
                    [2.0 / N / (N - 2), np.array([1.0 / N[-1]]), np.zeros(m)]
                )
                v2 = -v0[:-1] - v0[:0:-1]
                g0 = -np.ones(n)
                g0[length] += n
                g0[m] += n
                g = g0 / (n ** 2 - 1 + (n % 2))

                w = np.fft.ihfft(v2 + g)
                assert max(w.imag) < 1.0e-15
                w = w.real

                if n % 2 == 1:
                    weights = np.concatenate([w, w[::-1]])
                else:
                    weights = np.concatenate([w, w[len(w) - 2 :: -1]])
        weights = weights / 2.0
        return points, weights


class QuadFactory:
    # generates QuadBase class object
    @staticmethod
    def newQuad(quadtype="legendre_gauss"):
        if quadtype == "legendre_gauss":
            Q = LegendreQuad()
        if quadtype == "clenshaw_curtis":
            Q = ClenshawCurtis()
        if quadtype == "hermite_gauss":
            Q = HermiteQuad()
        return Q


class QuadRule:
    def __init__(self, x, w):
        self.x = x
        self.w = w
        self.n = len(w)
        if x.ndim == 1:
            self.dim = 1
            self.x = np.atleast_2d(x).T  # col vector
        else:
            self.dim = x.shape[1]
        assert len(x) == len(w), "x and w dont habe the same # of points"

    def __add__(self, other):
        assert self.dim == other.dim, "Dimensions do not match!"
        xnew = np.vstack([self.x, other.x])
        wnew = np.hstack([self.w, other.w])
        Qnew = QuadRule(xnew, wnew)
        return Qnew

    def __sub__(self, other):
        assert self.dim == other.dim, "Dimensions do not match!"
        xnew = np.vstack([self.x, other.x])
        wnew = np.hstack([self.w, -1 * other.w])
        Qnew = QuadRule(xnew, wnew)
        return Qnew

    def __mul__(self, other):
        # tensor product
        index_comb = list(itertools.product(range(self.n), range(other.n)))
        xnew = [np.concatenate([self.x[i[0]], other.x[i[1]]]) for i in index_comb]
        wnew = [self.w[i[0]] * other.w[i[1]] for i in index_comb]
        Qnew = QuadRule(np.array(xnew), np.array(wnew))
        return Qnew

    def copy(self):
        return QuadRule(self.x, self.w)


class QuadOps:
    @staticmethod
    def getMultiIndexLevel(level, ndim):
        """returns the multindices of order = level"""
        iup = 0
        nup_level = int(comb(ndim + level - 1, level))
        M = np.zeros((nup_level, ndim))
        if ndim == 1:
            M[0, 0] = level
        else:
            for first in range(level, -1, -1):
                theRest = QuadOps.getMultiIndexLevel(level - first, ndim - 1)

                for j in range(len(theRest)):
                    # print(iup,j)
                    M[iup, 0] = first
                    M[iup, 1:ndim] = theRest[j, 0 : ndim - 1]
                    iup += 1

        return M

    @staticmethod
    def compressRule(Q):
        # assert self.rule_ is not None, "Must set rule first."
        # convert numpy array to list of tuples
        xtuple = [tuple(xi) for xi in Q.x]
        # create a dictionary
        from collections import defaultdict

        dd = defaultdict(list)
        for ii, xi in enumerate(xtuple):
            dd[xi].append(Q.w[ii])
        # sum weights over keys
        for key in dd:
            dd[key] = np.sum(dd[key])
        x = np.array(list(dd.keys()))
        w = np.array([dd[key] for key in dd])
        x = x[np.abs(w) > 1e-12]
        w = w[np.abs(w) > 1e-12]
        return QuadRule(x, w)


# deprecated for sparse, but good for full tensor product grid
class QuadBuilder:
    def __init__(self, grid_type="sparse", order=2, quad_type="Legendre"):
        self.grid_type = grid_type
        self.quad_type = quad_type
        self.order = order
        self.ndim = None
        self.growth_rule = None

    def SetRule(self, ndim):
        self.ndim = ndim
        if self.grid_type == "full":
            self._full()
        if self.grid_type == "sparse":
            if self.quad_type == "legendre_gauss":
                self.growth_rule = 0
            self._sparse()
        return self

    def _full(self):
        # cannot do mixed quad yet. Easy if quad type takes in array
        quad_gen = QuadFactory.newQuad(self.quad_type)
        x, w = quad_gen.get1dQuad(nquad=self.order + 1)  # 0th order means 1 point
        q1d = QuadRule(x, w)
        qnew = q1d.copy()
        for i in range(1, self.ndim):
            qnew = qnew * q1d
        q = qnew.copy()
        self.rule_ = q

    def _sparse(self):
        for nlevel in range(-1, self.order):
            self._SetNextLevel2(nlevel)

    def _SetNextLevel(self, nlevel):
        nlevel += 1
        M = QuadOps.getMultiIndexLevel(nlevel, self.ndim)
        nM = M.shape[0]
        M_npts = np.zeros((nM, self.ndim))
        quad_gen = QuadFactory.newQuad(self.quad_type)
        for j in range(nM):
            Mj = M[j]  # jth row of the multiindexlevel
            # 1 if Mj == 0, 3 if 1, else (Mj_i+1)^2
            if self.growth_rule == 0:
                npts = 1 * (Mj == 0) + 3 * (Mj == 1) + ((Mj) ** 2 + 1) * (Mj > 1)
                npts_1 = 0 * (Mj == 0) + 1 * (Mj == 1) + ((Mj - 1) ** 2 + 1) * (Mj > 1)
            elif self.growth_rule == 1:
                npts = ((Mj + 1) ** 2 - 1) * (Mj > 1)
                npts_1 = (Mj ** 2 - 1) * (Mj > 1)
            npts = npts.astype("int")
            npts_1 = npts_1.astype("int")
            print(npts, npts_1)
            xw = [quad_gen.get1dQuad(nquad=int(n)) for n in list(npts)]
            rules = [QuadRule(xwi[0], xwi[1]) for xwi in xw]
            xw_1 = [quad_gen.get1dQuad(nquad=int(n_1)) for n_1 in list(npts_1)]
            rules_1 = [QuadRule(xwi_1[0], xwi_1[1]) for xwi_1 in xw_1]
            srules = []
            for ii in range(len(npts)):
                if npts_1[ii] > 0:
                    srules.append(rules[ii] - rules_1[ii])
                else:
                    srules.append(rules[ii])

            # multiply rules in srules
            r = srules[0].copy()
            for ri in srules[1:]:
                r = r * ri

            if j == 0:
                rule_level = r.copy()
            else:
                rule_level = r + rule_cur
            rule_cur = rule_level.copy()
        # pdb.set_trace()
        if nlevel == 0:
            rule_total = rule_level.copy()
        else:
            rule_total = self.rule_ + rule_level
        self.rule_ = rule_total.copy()
        return self

    def _SetNextLevel2(self, nlevel):
        nlevel += 1
        M = QuadOps.getMultiIndexLevel(nlevel, self.ndim)
        self.M = M
        # nM = M.shape[0]
        M_npts = np.zeros(M.shape)
        quad_gen = QuadFactory.newQuad(self.quad_type)
        for j in range(len(M)):
            rules = []
            rules_1 = []
            srules = []
            Mj = M[j]
            for id in range(self.ndim):
                if M[j, id] == 0:
                    npts = 1
                    npts_1 = 0
                elif M[j, id] == 1:
                    npts = 3
                    npts_1 = 1
                else:
                    npts = int((M[j, id]) ** 2) + 1
                    npts_1 = int((M[j, id] - 1) ** 2) + 1

                x, w = quad_gen.get1dQuad(nquad=npts)
                rule = QuadRule(x, w)
                if npts_1 > 0:
                    x_1, w_1 = quad_gen.get1dQuad(nquad=npts_1)
                    rule_1 = QuadRule(x_1, w_1)
                    srule = rule - rule_1
                else:
                    srule = rule.copy()
                srules.append(srule)
                # end of id iterator
            # multiple rule
            rule_temp = srules[0].copy()
            for s in srules[1:]:
                rule_temp = rule_temp * s
            # rule_temp = srules[0]*srules[1]
            if j == 0:
                rule_level = rule_temp.copy()
            else:
                rule_level = rule_cur + rule_temp
            # pdb.set_trace()
            rule_cur = rule_level.copy()

            # end of j iterator

        if nlevel == 0:
            rule_total = rule_level.copy()
        else:
            rule_total = self.rule_ + rule_level
        self.rule_ = rule_total.copy()
        # test = np.unique(self.rule_.x,axis=0)
        # ind = []
        # for i in range(len(test)):
        # 	ind = np.array_equal(test[0],self.rule_.x[i])
        # 	if ind == True:
        # 		print(self.rule_.w[i])
        # ipdb.set_trace()
        self.rule_ = QuadOps.compressRule(self.rule_)
        return self


def construct_lookup(orders, dim, rules="gaussian"):
    """
    Create abscissas and weights look-up table so values do not need to be
    re-calculatated on the fly.

        This code has been adapted from Chaospy (https://github.com/jonathf/chaospy) using custom multiindex generation and circumventing the use of external libraries like numpoly.
    """
    if isinstance(rules, str):
        rules = (rules,) * dim
    if isinstance(orders, int):
        orders = orders * np.ones(dim, dtype=int)
    x_lookup = []
    w_lookup = []
    if rules[0] == "gaussian" or "legendre_gauss":
        Q = QuadFactory.newQuad("legendre_gauss")
        growth = False
    if rules[0] == "clenshaw_curtis":
        Q = QuadFactory.newQuad("clenshaw_curtis")
        growth = True
    # if rules[0] == 'hermite':
    # Q = QuadFactory.newQuad("hermite_gauss")
    for max_order, rule in zip(orders, rules):
        x_lookup.append([])
        w_lookup.append([])
        for order in range(max_order + 1):
            if growth == True:
                if order == 0:
                    order_adj = 1
                else:
                    order_adj = 2 ** (order) + 1
            else:
                order_adj = order + 1
            abscissas, weights = Q.get1dQuad(order_adj)
            abscissas[np.abs(abscissas) <= 1e-16] = 0.0
            weights[np.abs(weights) <= 1e-16] = 0.0
            x_lookup[-1].append(abscissas)
            w_lookup[-1].append(weights)
    return x_lookup, w_lookup


def construct_collection(orders, dim, x_lookup, w_lookup):
    """Create a collection of {abscissa: weight} key-value pairs.

    This code has been adapted from Chaospy (https://github.com/jonathf/chaospy) using custom multiindex generation and circumventing the use of external libraries like numpoly.
    """
    if isinstance(orders, int):
        orders = orders * np.ones(dim, dtype=int)
    order = np.min(orders)
    skew = orders - order

    mi = []
    for ilevel in range(order - dim + 1, order + 1):
        mi.append(QuadOps.getMultiIndexLevel(ilevel, dim))
    indices = np.vstack(mi).astype("int")
    coeffs = np.sum(indices, -1)
    coeffs = (2 * ((order - coeffs + 1) % 2) - 1) * comb(dim - 1, order - coeffs)

    collection = defaultdict(float)
    for bidx, coeff in zip(indices + skew, coeffs.tolist()):
        abscissas = [value[idx] for idx, value in zip(bidx, x_lookup)]
        weights = [value[idx] for idx, value in zip(bidx, w_lookup)]
        for abscissa, weight in zip(product(*abscissas), product(*weights)):
            collection[abscissa] += np.prod(weight) * coeff

    return collection


from collections import defaultdict
from itertools import product

# for gaussian, since quad points are as high accuracy as chaospy, slight differences


class QuadGen(object):
    def __init__(self, dim, order=2, rule="legendre_gauss", sparse=True):
        self.dim = dim
        self.order = order
        self.rule = rule
        self.sparse = sparse

    def run(self, order=None):
        if order is None:
            order = self.order
        if self.sparse == False:
            Qtemp = QuadBuilder(order=order, grid_type="full", quad_type=self.rule)
            Qtemp.SetRule(ndim=self.dim)
            x, w = Qtemp.rule_.x, Qtemp.rule_.w
        elif self.sparse == True:
            # create initial lookup table of 1d quadratures
            self.x_lookup, self.w_lookup = construct_lookup(
                orders=order, dim=self.dim, rules=self.rule
            )
            self.collection = construct_collection(
                orders=order,
                dim=self.dim,
                x_lookup=self.x_lookup,
                w_lookup=self.w_lookup,
            )
            # self.collection = _construct_collection_custom(self.order, self.dim, self.x_lookup, self.w_lookup)
            # scale and renormalize to over [0,1] and unit weight
            x = sorted(self.collection)
            w = np.array([self.collection[key] for key in x])
            x = np.array(x)  # each row is a data point
            assert (
                np.abs(np.sum(w) - 1.0)
            ) <= 1e-12, "Weights are not normalized on [0,1]."
        x = 0.5 * (x + 1)
        assert np.all(x >= 0.0) and np.all(
            x <= 1.0
        ), "Points are outside the range [0,1]^d"
        self.points, self.weights = x, w
        return self.points, self.weights
